"""
federation/resolver.py
ActivityPub resolver implementation.
"""

from collections.abc import AsyncIterator
from http import HTTPStatus
from json import JSONDecodeError
import json
from logging import Logger
from typing import Any, cast
from urllib.parse import urlparse

import httpx
from pydantic import SecretStr
from urllib3.util import Url

from phederation.cache import BaseCache
from phederation.cache.base import WithCache, with_cache
from phederation.federation.instance import APInstance
from phederation.models import (
    APAccount,
    APActivity,
    APActor,
    APCollection,
    APCollectionPage,
    APObject,
    APOrderedCollection,
    APOrderedCollectionPage,
    APPrivateKey,
    dereference,
    object_from_type,
)
from phederation.models.objects import dereference_or_raise
from phederation.models.proofs import DataIntegrityProof
from phederation.storage import StorageBackend
from phederation.utils import ObjectId, PUBLIC_URLS, UrlType, validate_object_id
from phederation.utils.base import (
    AccessType,
    ActivityType,
    NodeInfo,
    ObjectType,
    ResolverResult,
    assemble_id_url,
    collection_id_from_name,
    username_from_actor_id,
)
from phederation.utils.exceptions import (
    ActivityPubException,
    AuthorizationError,
    ResolverError,
)
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings
from phederation.utils.version import PHEDERATION_USERAGENT


class ActivityPubResolver(WithCache):
    """Resolves ActivityPub resources.

    Features:
     - Resolve objects and activities over the network.
     - Local resolution inside an instance, from storage or a dictionary."""

    def __init__(
        self,
        settings: PhedSettings,
        storage: StorageBackend | None = None,
        instance: APInstance | None = None,
        cache: BaseCache | None = None,
        timeout: int = 3,
    ):
        """Create new resolver.

        Args:
            settings (PhedSettings): The settings for the instance, as much as they are available.
            local_domain (str | None, optional): base_url for the local instance. Defaults to None.
            storage (StorageBackend | None, optional): Storage backend, if not in c2s mode. Defaults to None.
            instance (APInstance | None, optional): Instance, if not in c2s mode. Defaults to None.
            cache (BaseCache | None, optional): Optional cache for actor, object, and collection data. Defaults to None.
            timeout (int, optional): seconds until timeout for resolver requests. Defaults to 3.
        """
        self.settings: PhedSettings = settings
        self.cache: BaseCache | None = cache
        self.timeout: int = timeout
        self.storage: StorageBackend | None = storage
        self.instance: APInstance | None = instance
        self.local_domain: str | None = settings.domain.base_url
        # this can be set by the test suite to replace sending actual http requests and instead use the test_client
        self.requests: httpx.AsyncClient = httpx.AsyncClient()
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

    def is_local(self, url: ObjectId) -> bool:
        return self.local_domain is not None and str(url).startswith(self.local_domain)

    async def try_resolve_from_storage(
        self, url: ObjectId, url_type: UrlType
    ) -> APObject | APActivity | APCollection | APOrderedCollection | APActor | APPrivateKey | DataIntegrityProof | NodeInfo | None:
        """Tries to resolve an object with the given id from the local storage.

        Note that this does **not** attach a DataIntegrityProof to the object here, even if it is required by settings.
        Currently, only "/users", "/activities" and "/objects" paths are resolved,
        and only if they directly point to an actor, activity, or object (not collections of them).

        Args:
            url (ObjectId): Id of the object to resolve.
            url_type (UrlType): the type of the url to resolve (Users, Activities, ...). This is required because not all urls make it obvious what objects they are resolving.

        Returns:
            APObject | APActivity | APCollection | APOrderedCollection | APActor | APPrivateKey | DataIntegrityProof | NodeInfo | None: Object resolved from storage.
        """
        # do *not* check if the url is local, so that we can properly resolve cached remote objects
        # if self.is_local(url):
        if self.storage:
            try:
                if url_type == UrlType.Users:
                    return await self.storage.actor.read(id=url)
                elif url_type == UrlType.Activities:
                    return await self.storage.activity.read(id=url)
                elif url_type == UrlType.Objects:
                    return await self.storage.object.read(id=url)
                elif url_type == UrlType.Keys:
                    return await self.storage.key.read(id=url)
                elif url_type == UrlType.Proofs:
                    return await self.storage.proof.read(id=url)
                elif url_type == UrlType.NodeInfo:
                    return await self.storage.node_info.read(id=url)
            except ActivityPubException as e:
                self.logger.debug(e.message)
            except Exception as e:
                self.logger.debug(f"Could not resolve {url} from local storage; {e}")

    @with_cache(__name__)
    async def resolve(
        self, url: ObjectId, url_type: UrlType, as_bytes: bool = False, access_token: SecretStr | None = None
    ) -> (
        AsyncIterator[bytes]
        | APObject
        | APActivity
        | APCollection
        | APOrderedCollection
        | APActor
        | APAccount
        | APPrivateKey
        | DataIntegrityProof
        | NodeInfo
    ):

        result = await self.try_resolve_from_storage(url, url_type=url_type)
        if result:
            if (
                isinstance(result, DataIntegrityProof)
                or isinstance(result, NodeInfo)
                or AccessType.from_visibility(result.visibility) == AccessType.PUBLIC
            ):
                return result
            elif access_token:
                # do not expose the object, and try to resolve it over the network
                result = None
            else:
                raise AuthorizationError(f"Cannot resolve, not authorized (result.visibility={result.visibility})")
        if not result:
            if self.instance and await self.instance.is_blocked_instance(url):
                raise AuthorizationError(message=f"Cannot resolve, instance for url {url} is blocked")

            headers = {
                "Accept": 'application/ld+json; profile="https://www.w3.org/ns/activitystreams"',
                "User-Agent": PHEDERATION_USERAGENT,
            }

            # only use the token if the request is on the local server!
            # Otherwise, the access token for the user gets sent to servers on the network, a security leak.
            if access_token:
                if self.is_local(url):
                    headers["Authorization"] = f"Bearer {access_token.get_secret_value()}"
                    headers["WWW-Authenticate"] = "Bearer"
                else:
                    self.logger.warning("Trying to access remote resource with user access token. This is a security issue - will not use the token.")

            try:
                if self.is_local(url) and self.instance:
                    url_parsed = urlparse(url)
                    url_local = Url(
                        scheme="http",
                        host=self.settings.domain.docker_url,
                        port=self.settings.domain.docker_port,
                        fragment=url_parsed.fragment if len(url_parsed.fragment) > 0 else None,
                        query=url_parsed.query if len(url_parsed.query) > 0 else None,
                        path=url_parsed.path if len(url_parsed.path) > 0 else None,
                    ).url
                    url = url_local
                response = await self.requests.get(str(url), headers=headers, timeout=self.timeout)
            except httpx.ConnectError as e:
                result = ResolverResult(content=None, error_message=f"Connection error for {url}", status_code=None)
                raise ResolverError(f"Failed to resolve {url}, error: {e}", response=result)

            if response.status_code != HTTPStatus.OK:
                result = self.result_from_response(response, message=f"Failed to resolve url {url}")
                raise ResolverError(message=f"Failed to resolve url {url}: {response}, {response.content}", response=result)

            try:
                if as_bytes:
                    # response_bytes = response.iter_bytes()  TODO: use iter_bytes instead, to slow down download
                    result = response.aiter_bytes()
                else:
                    response_dict = cast(dict[str, Any], response.json())
                    result = self.resolve_from_dict(response_dict)
            except JSONDecodeError as e:
                result = self.result_from_response(response, message=f"JSONDecodeError for response from url {url}")
                raise ResolverError(f"ActivityPubResolver: JSONDecodeError for response {response}: {e}.", response=result)

        return result

    def result_from_response(self, response: httpx.Response, message: str | None = None):
        content_str = response.content.decode(errors="ignore")
        try:
            content_dict = cast(dict[str, str], json.loads(content_str))
        except JSONDecodeError:
            content_dict = {"bytes": content_str}
        return ResolverResult(
            content=content_dict,
            error_message=message,
            status_code=response.status_code,
        )

    async def resolve_media(self, file_id: ObjectId, access_token: SecretStr | None = None) -> AsyncIterator[bytes]:
        """Resolves the content of the file.

        Returns an async iterator to download the file content.

        Args:
            file_id (ObjectId): The url to download the file.
            access_token (SecretStr | None, optional): Access token of the user that can access the file. Defaults to None.

        Raises:
            ResolverError: If the file cannot be accessed.

        Returns:
            AsyncIterator[bytes]: Iterator for the bytes of the file.
        """
        self.logger.info(f"Resolving media with id {file_id} over network")

        result = await self.resolve(file_id, url_type=UrlType.Media, as_bytes=True, access_token=access_token)
        if not isinstance(result, AsyncIterator):
            raise ResolverError(f"Cannot download media from '{file_id}', is not a bytes array.")
        return result

    async def resolve_actor(self, actor_id: ObjectId | str | dict[str, Any] | APObject | None, access_token: SecretStr | None = None) -> APActor:
        """
        Resolve an actor by ID or account.

        This will try to resolve in the following order.
          1. From cache,
          2. From local storage (even for remote actors),
          3. From remote network / instance.
        This order ensures that actors are cached locally and not a lot of fetch requests are required.

        Args:
            actor_id: Actor ID or account (user@domain)

        Returns:
            Actor data or None if not found
        """
        if isinstance(actor_id, APObject):
            actor_id = actor_id.id
        if not actor_id:
            raise ResolverError("Cannot resolve actor_id, is None")
        if isinstance(actor_id, dict):
            actor = self.resolve_from_dict(actor_id)
            if not isinstance(actor, APActor):
                raise ResolverError(f"Failed to deserialize APActor (was type '{type(actor)}'): {actor.serialize()}")
            return actor
        if isinstance(actor_id, APActor):
            # nothing to do, just return it
            return actor_id

        actor = await self.resolve(actor_id, url_type=UrlType.Users, access_token=access_token)

        if not isinstance(actor, APActor):
            raise ResolverError(f"ActivityPubResolver: Failed to deserialize actor {actor_id}, type: '{type(actor).__name__}'.")

        return actor

    async def resolve_account(self, actor_id: ObjectId, access_token: SecretStr | None = None):
        parsed_url = urlparse(actor_id)
        base_url = validate_object_id(Url(scheme=parsed_url.scheme, host=parsed_url.netloc).url)
        username = username_from_actor_id(actor_id)
        account_url = assemble_id_url(UrlType.Accounts, base_url=base_url, primary=username)
        account = await self.resolve_object(account_url, url_type=UrlType.Accounts, access_token=access_token)

        if not isinstance(account, APAccount):
            raise ResolverError(f"ActivityPubResolver: Failed to deserialize account for {actor_id}, type: '{type(account).__name__}'.")
        return account

    def resolve_from_dict(
        self, object_dict: dict[str, Any]
    ) -> APObject | APActivity | APCollection | APOrderedCollection | APActor | APAccount | APPrivateKey | DataIntegrityProof | NodeInfo:
        object_type = object_dict.get("type", None)
        if not object_type:
            object_type = "Object"
        if not isinstance(object_type, str):
            raise ResolverError("Resolving type from object_dict did not result in a string")

        obj = object_from_type(object_dict)

        if (
            not isinstance(obj, APObject)
            and not isinstance(obj, APPrivateKey)
            and not isinstance(obj, DataIntegrityProof)
            and not isinstance(obj, NodeInfo)
        ):
            raise ResolverError(f"ActivityPubResolver: Cannot deserialize object (type {(type(obj))}): {object_dict}.")
        return obj

    async def resolve_nodeinfo(self, id: ObjectId | str | dict[str, Any] | NodeInfo | None, access_token: SecretStr | None = None) -> NodeInfo:
        if not id:
            raise ResolverError("Cannot resolve NodeInfo, id is None")
        # nothing to do, just return it
        if isinstance(id, NodeInfo):
            return id

        if isinstance(id, dict):
            obj = self.resolve_from_dict(id)
        else:
            obj = await self.resolve(id, url_type=UrlType.NodeInfo, access_token=access_token)
        if not isinstance(obj, NodeInfo):
            raise ResolverError("Object to resolve must be a NodeInfo. Use resolve_object, resolve_proof, or resolve_media for other types of url.")

        return obj

    async def resolve_proof(
        self, proof_id: ObjectId | str | dict[str, Any] | DataIntegrityProof | None, access_token: SecretStr | None = None
    ) -> DataIntegrityProof:
        if not proof_id:
            raise ResolverError("Cannot resolve object, id is None")
        # nothing to do, just return it
        if isinstance(proof_id, DataIntegrityProof):
            return proof_id

        if isinstance(proof_id, dict):
            obj = self.resolve_from_dict(proof_id)
        else:
            obj = await self.resolve(proof_id, url_type=UrlType.Proofs, access_token=access_token)
        if not isinstance(obj, DataIntegrityProof):
            raise ResolverError("Object to resolve must be a DataIntegrityProof. Use resolve_object or resolve_media for other types of url.")

        return obj

    async def resolve_object(
        self, object_id: ObjectId | str | dict[str, Any] | APObject | None, access_token: SecretStr | None = None, url_type: UrlType = UrlType.Unknown
    ) -> APObject | APActivity | APCollection | APOrderedCollection | APActor | APAccount | APPrivateKey:
        """
        Resolve an object by ID.

        Args:
            object_id: Object ID

        Returns:
            Object data or None if not found
        """
        if not object_id:
            raise ResolverError("Cannot resolve object, id is None")
        # nothing to do, just return it
        if isinstance(object_id, APObject) or isinstance(object_id, APPrivateKey):
            return object_id

        if isinstance(object_id, dict):
            obj = self.resolve_from_dict(object_id)
        else:
            if isinstance(object_id, APObject):
                object_type = dereference_or_raise(object_id, "type")
                url_type = UrlType.from_object_type(object_type)
            obj = await self.resolve(object_id, url_type=url_type, access_token=access_token)
        if isinstance(obj, AsyncIterator):
            raise ResolverError("Object to resolve must not be a media file, use resolve_media")
        if isinstance(obj, DataIntegrityProof):
            raise ResolverError("Object to resolve must not be a DataIntegrityProof, use resolve_proof")
        if isinstance(obj, NodeInfo):
            raise ResolverError("Object to resolve must not be a NodeInfo, use resolve_nodeinfo")

        return obj

    async def resolve_activity(
        self, activity_id: ObjectId | str | dict[str, Any] | APActivity | APObject | None, access_token: SecretStr | None = None
    ) -> APActivity:
        """
        Resolve an activity by ID.

        Args:
            activity_id: Activity ID

        Returns:
            Activity data or None if not found
        """
        activity = await self.resolve_object(activity_id, url_type=UrlType.Activities, access_token=access_token)
        if not isinstance(activity, APActivity):
            raise ResolverError(f"Failed to deserialize APActivity (is type {type(activity)}): {activity.serialize() if activity else 'None'}")
        return activity

    async def resolve_items(self, collection_id: ObjectId, max_pages: int = 100, access_token: SecretStr | None = None) -> list[ObjectId]:
        """Resolve all items in a paginated collection.

        This includes all pages and all contained items, until max_pages is reached.
        The items themselves are still represented as ObjectId alone, not fully resolved.

        Args:
            collection_id (ObjectId): id of the collection to resolve all items from.
            max_pages (int, optional): maximum number of pages to resolve. Defaults to 100.
            access_token (str | None, optional): token for authentication. Defaults to None.

        Raises:
            ResolverError: If the collection or requested page cannot be found.

        Returns:
            list[ObjectId]: List of all ids of the items.
        """
        collection_obj = await self.resolve_object(collection_id, url_type=UrlType.Collections, access_token=access_token)
        if not isinstance(collection_obj, APCollection):
            raise ResolverError("Could not resolve APCollection")

        resolved_items: list[ObjectId] = []
        total_items = collection_obj.total_items or 0

        if total_items == 0:
            return []

        items = collection_obj.items or []
        if isinstance(collection_obj, APOrderedCollection):
            items = collection_obj.ordered_items or []

        # the next page to visit depends on if the current collection is a generic collection or already a page
        next_page = dereference(collection_obj, "first")
        if collection_obj.type in [ObjectType.COLLECTION_PAGE.value, ObjectType.ORDERED_COLLECTION_PAGE.value]:
            next_page = dereference(collection_obj, "next")

        visited_pages: list[ObjectId] = []
        pages_to_visit: list[APCollection] = []

        if collection_obj.id:
            visited_pages.append(collection_obj.id)

        # at least try to visit the first page, if we visit any pages at all. From that page we will then visit all pages.
        if max_pages > 0 and next_page:
            next_page = await self.resolve_object(next_page, access_token=access_token)
            if isinstance(next_page, APCollection):
                pages_to_visit.append(next_page)

        for _ in range(max_pages):
            for item in items:
                if isinstance(item, str):
                    resolved_items.append(item)
                    continue
                item_id = dereference(item, "id")
                if not item_id:
                    raise ResolverError(f"Could not resolve id of item {item} in collection {collection_id}. Aborting.")

                # if we have visited the given item before, do not add its items again
                if item_id in visited_pages:
                    continue

                item_obj = await self.resolve_object(object_id=item_id, access_token=access_token)

                if any(
                    isinstance(item_obj, item_type) for item_type in [APCollection, APCollectionPage, APOrderedCollection, APOrderedCollectionPage]
                ):
                    # resolve the items directly, do not paginate here
                    resolved_items.extend(await self.resolve_items(item_id, max_pages=1, access_token=access_token))
                    visited_pages.append(item_id)
                    next_page = dereference(item, "next")
                    if next_page:
                        next_page = await self.resolve_object(next_page, access_token=access_token)
                        if isinstance(next_page, APCollection):
                            pages_to_visit.append(next_page)
                else:
                    resolved_items.append(item_id)
            items = pages_to_visit
            pages_to_visit = []

            if len(resolved_items) >= total_items:
                break

        return resolved_items

    async def resolve_collection_id(self, collection: str, actor_id: ObjectId, access_token: SecretStr | None = None) -> ObjectId:
        """Resolves only the id of a collection of the given actor or object.

        Args:
            actor_id (ObjectId): id of the actor.
            collection (str): name of the collection, e.g. "follower", "inbox", "outbox", "shares" (for objects), ...

        Returns:
            ObjectId: id of the collection

        Raises:
            ResolverError: If the collection cannot be found
        """
        # TODO: fix that we have to use "resolve_object" here because the actor_id may not actually be from an actor.
        response_obj = await self.resolve_object(object_id=actor_id, access_token=access_token)

        if not hasattr(response_obj, collection):
            raise ResolverError(
                f"ActivityPubResolver: actor/object {actor_id} does not have a collection '{collection}'. Dictionary: {response_obj}."
            )

        # now resolve the collection id using the link in the response
        collection_id = dereference(response_obj, key=collection)

        if not collection_id:
            # this sually means the collection is empty, so "None" was returned, not serialized.
            collection_id = collection_id_from_name(id=actor_id, name=collection)

        return collection_id

    async def resolve_collection(
        self, collection: str, actor_id: ObjectId, access_token: SecretStr | None = None
    ) -> APCollection | APOrderedCollection:
        """Resolves a collection of the given actor or object.
        The items or ordered_items property may not be fully resolved by this method, because it may contain a large number of objects.
        Pagination must be used to fully resolve it.

        Args:
            actor_id (ObjectId): id of the actor.
            collection (str): name of the collection, e.g. "follower", "inbox", "outbox", "shares" (for objects), ...

        Returns:
            APCollection | APOrderedCollection, depending on the type of collection.
        """
        # now resolve the collection using the link in the response
        collection_id = await self.resolve_collection_id(collection=collection, actor_id=actor_id, access_token=access_token)

        collection_obj = await self.resolve_object(object_id=collection_id, access_token=access_token)
        if not collection_obj:
            raise ResolverError(f"Collection with id '{collection_id}' was not accessible.")
        if not collection_obj.id:  # the case for collections
            collection_obj.id = collection_id

        if isinstance(collection_obj, APCollection):
            return collection_obj
        else:
            raise ResolverError(
                f"Collection with id '{collection_id}' did not resolve to an APCollection(Page) or APOrderedCollection(Page). Was: {collection_obj.serialize()}"
            )

    async def resolve_inboxes(self, recipients: list[ObjectId], depth: int = 3, access_token: SecretStr | None = None) -> list[ObjectId]:
        if depth < 0:
            self.logger.debug(f"ActivityPubResolver: resolution depth for recipient list reached.")
            return recipients

        # add instance actor inbox to recipients if this should be sent to Public
        recipients_without_public: list[ObjectId] = []
        for recipient_id in recipients:
            if recipient_id in PUBLIC_URLS:
                # add actors of federated instances, i.e. send to "everybody"
                if self.instance:
                    federated_inboxes = await self.instance.get_federated_instance_actors(access=AccessType.PRIVATE)
                    recipients.extend(federated_inboxes)
            else:
                recipients_without_public.append(recipient_id)
        recipients = recipients_without_public

        # deduplicate list
        recipients = list(set(recipients))

        # resolve inboxes of all actors
        resolved_list: list[ObjectId] = []
        for recipient_id in recipients:
            # try to resolve the id as a general object
            result_obj = await self.resolve_object(object_id=recipient_id, access_token=access_token)
            if isinstance(result_obj, APCollection) or isinstance(result_obj, APCollectionPage):
                if not result_obj.id:
                    raise ResolverError("Collection does not have an id")
                items = await self.resolve_items(collection_id=result_obj.id, access_token=access_token)

                items_resolved: list[ObjectId] = []
                for item in items:
                    item_id = dereference(item, key="id") if isinstance(item, dict) else item
                    if item_id:
                        items_resolved.append(item_id)

                # resolve all groups in the new list, too
                resolved_list += await self.resolve_inboxes(items_resolved, depth=depth - 1, access_token=access_token)
            elif isinstance(result_obj, APActivity) and result_obj.type == ActivityType.FOLLOW.value:
                actor_id = dereference(result_obj, key="actor")
                if not actor_id:
                    raise ResolverError("actor_id is none in resolved APFollow object")
                if self.instance and await self.instance.is_banned_actor(actor_id):
                    self.logger.debug(f"Not delivering to banned actor: {actor_id}")
                else:
                    actor = await self.resolve_actor(actor_id, access_token=access_token)
                    resolved_list.append(actor.inbox)
            elif isinstance(result_obj, APActor):
                if self.instance and await self.instance.is_banned_actor(result_obj.id):
                    self.logger.debug(f"Not delivering to banned actor: {result_obj.id}")
                else:
                    resolved_list.append(result_obj.inbox)
            else:
                self.logger.error(f"Cannot resolve recipient: {recipient_id} (wrong type: {type(result_obj)}). Failing silently.")
        return resolved_list

    async def resolve_access_type(self, object: APObject, actor_requesting_access: ObjectId | None):
        """Returns the given object if the given user is allowed to access it.

        Args:
            object (APObject): object to return.
            userinfo_requesting_access (UserInfo | None): Description of the actor trying to access the object.

        Returns:
            APObject | None: If the object is accessible by the user, else None.
        """
        if object.visibility.lower() == "public":
            return object

        actors_with_access = [dereference(object, "attributed_to"), dereference(object, "actor")]
        actors_with_access += await object.get_recipients()
        actors_with_access = list(set(actors_with_access))
        actors_with_access = [a for a in actors_with_access if a is not None]

        object_access = AccessType.from_visibility(object.visibility)
        if (
            (object_access == AccessType.PRIVATE or object_access == AccessType.FOLLOWER)
            and actor_requesting_access
            and actor_requesting_access in actors_with_access
        ):
            return object
        # TODO: handle follower lists
        # if attributed_to_id and object.visibility == AccessType.FOLLOWER.value and actor_requesting_access and access_token:
        #     collection_id = collection_id_from_name(id=attributed_to_id, name=UrlType.Follower.value)
        #     follower = await self.resolve_items(collection_id=collection_id, access_token=access_token)
        #     if actor_requesting_access in follower:
        #         return object



"""Accept activity handler."""

from datetime import datetime
from typing import override

from phederation.handlers.migrate import MigrationHandler
from phederation.models import dereference
from phederation.models.activities import APActivity, APFollow, APMigrate
from phederation.models.collections import ValidCollection
from phederation.models.objects import dereference_or_raise
from phederation.utils import ObjectId
from phederation.utils.base import AccessType, ActivityType, collection_id_from_name
from phederation.utils.exceptions import HandlerError, ValidationError

from .base import ActivityHandler


class AcceptHandler(ActivityHandler):
    """Handles Accept activities."""

    ACCEPTABLE_TYPES: list[str] = [ActivityType.FOLLOW.value, ActivityType.MIGRATE.value]

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Accept activity."""
        if activity.type != "Accept":
            raise ValidationError("Invalid activity type")

        activity_to_accept = await self.resolver.resolve_activity(activity.object)
        if not activity_to_accept:
            raise ValidationError("Accept must have an object, set to the activity to accept")

        # Check if object is a dict with type
        if activity_to_accept.type not in self.ACCEPTABLE_TYPES:
            raise ValidationError(f"Cannot accept activity type: {activity_to_accept.type}")

        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """Handle Accept activity."""
        actor_id = dereference(data=activity, key="actor")
        activity_to_accept = await self.resolver.resolve_activity(dereference(activity, "object"))

        if not actor_id:
            raise HandlerError(f"Actor not found: actor_id is None for activity {activity}.")

        activity_actor_id = dereference_or_raise(data=activity_to_accept, key="actor")

        # Handle acceptance
        access = AccessType.from_visibility(activity.visibility)
        if activity_to_accept.type == ActivityType.FOLLOW.value:
            follow_activity_id = dereference_or_raise(data=activity_to_accept, key="id")
            following_id = dereference_or_raise(data=activity_to_accept, key="object")
            # Verify authorization for APFollow
            if following_id != actor_id:
                raise HandlerError(f"Unauthorized: can only accept APFollow activities targeting self. {following_id} is not {actor_id}")
            await self._handle_accept_follow(follower=activity_actor_id, following=following_id, follow_activity_id=follow_activity_id, access=access)
        elif activity_to_accept.type == ActivityType.MIGRATE.value and isinstance(activity_to_accept, APMigrate):
            target_actor_id = dereference(data=activity_to_accept, key="actor_to")
            # Verify authorization for APFollow
            if target_actor_id != actor_id:
                raise HandlerError(
                    f"Unauthorized: can only accept APMigrate activities targeting self. {target_actor_id} is not {actor_id}; activity={activity_to_accept.serialize()}"
                )
            await MigrationHandler.accept_migrate_on_new_instance(handler=self, activity=activity_to_accept)

        # Notify original actor: not necessary, because it should be in the "to" field of the activity already
        return activity

    async def _handle_accept_follow(self, follower: ObjectId, following: ObjectId, follow_activity_id: ObjectId, access: AccessType) -> None:
        """Handle accepting a Follow."""
        _ = await self.storage.follow.upsert(
            id=follow_activity_id, data=APFollow(id=follow_activity_id, actor=follower, object=following, accepted=True, accepted_at=datetime.now())
        )
        collection_follow = collection_id_from_name(id=following, name=ValidCollection.Followers.value)
        _ = await self.collections.add_to_collection(collection_id=collection_follow, items=follow_activity_id, access=access, add_only_once=True)

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        """Process inbox for APAccept.

        Returns:
            Optional[str]: Follow id stored with the accept processed, if it was created anew.
        """
        # Resolve the activity to accept
        accepted_activity_id = dereference_or_raise(activity, key="object")
        accepted_activity = await self.resolver.resolve_activity(accepted_activity_id)

        actor_id = dereference_or_raise(accepted_activity, "actor")
        target_id = dereference_or_raise(data=accepted_activity, key="object")

        if accepted_activity.type == ActivityType.FOLLOW.value:
            self.logger.info(f"Accepting APFollow in inbox of actor {actor_id}")
            # If the object is an activity of type Follow, the server should also check:
            #     That the actor of the Accept activity is not already in the recipient's following collection.
            if await self.storage.is_following(follower=actor_id, following=target_id, check_accepted=True):
                self.logger.info("Follow already in status ACCEPT locally.")
                return None
            # TODO: Check That the Follow activity has not been undone by an Undo activity. See ActivityPub/Primer/Undo activity for more information.

            # If these checks pass, then the server should:
            #     Add the actor of the Accept activity to the actor of the Follow activity's following.
            _ = await self.storage.follow.upsert(
                id=accepted_activity_id,
                data=APFollow(id=accepted_activity_id, actor=actor_id, object=target_id, accepted=True, accepted_at=datetime.now()),
            )
            collection_follow = collection_id_from_name(id=actor_id, name=ValidCollection.Following.value)
            if not await self.collections.contains(collection_id=collection_follow, item=accepted_activity_id):
                access = AccessType.from_visibility(accepted_activity.visibility)
                _ = await self.collections.add_to_collection(collection_id=collection_follow, items=accepted_activity_id, access=access)
        elif accepted_activity.type == ActivityType.MIGRATE.value and isinstance(accepted_activity, APMigrate):
            self.logger.info(f"Accepting APMigrate in inbox of actor {actor_id}")
            await MigrationHandler.accept_migrate_on_previous_instance(handler=self, activity=accepted_activity)
        else:
            self.logger.warning(f"Could not process activity {accepted_activity.type} in inbox (incorrect type)")

        # TODO: If the server is keeping caches of remote followers collects, add the actor of the Follow activity to the actor of the Accept activity's following.
        return activity.id



from abc import ABC, abstractmethod
from logging import Logger

from fastapi import FastAPI

from phederation.federation import ActivityPubServer
from phederation.federation.middleware import ActivityPubMiddleware
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings


class BaseAPI(ABC):
    """Base class for all API implementations."""

    async def initialize(self, app: FastAPI, settings: PhedSettings, middleware: ActivityPubMiddleware, server: ActivityPubServer) -> None:
        self.app: FastAPI = app
        self.settings: PhedSettings = settings
        self.server: ActivityPubServer = server
        self.middleware: ActivityPubMiddleware = middleware
        self.logger: Logger = configure_logger(
            f"{".".join(__name__.split('.')[:-1])}.{type(self).__name__.lower()}", prefix=self.settings.federation.logging_prefix
        )

    @abstractmethod
    def version(self, major: bool = True, minor: bool = True, patch: bool = True) -> str:
        pass



"""
keys.py
This module defines key objects used in ActivityPub.
"""

from typing import ClassVar, override

from pydantic import ConfigDict, Field

from phederation.models.errors import BaseError
from phederation.models.objects import APObject
from phederation.utils.base import APDateTime


class ErrorMessageKeyNotFound(BaseError):
    message: str = "Key not found"
    description: str = "The key with the given id could not be found on the instance."


class ErrorMessageKeyRestricted(BaseError):
    message: str = "Key has restricted visibility."
    description: str = "The key with the given id has restricted visibility."


class APPublicKey(APObject):
    """ActivityPub public key class."""

    type: str = Field(default="Key", description="This is a key.")
    public_key_pem: str | None = Field(default=None, description="The PEM string of the public key.")
    created_at: APDateTime | None = Field(default=None, description="Created time")
    expires_at: APDateTime | None = Field(default=None, description="Expiry time")

    def is_expired(self, now: APDateTime):
        if self.expires_at and now > self.expires_at:
            return True
        else:
            return False

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "id": "https://example.org/keys/odhfsd-sdfsdf-sdrqwe434-adasdasd-12311eawed",
                    "type": "Key",
                    "publicKeyPem": "<pem string>",
                }
            ]
        }
    )

    @override
    def __str__(self):
        return str(self.id)


class APPrivateKey(APPublicKey):
    """ActivityPub private key class."""

    private_key_pem: str | None = Field(default=None, description="The PEM string of the private key.")

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "id": "https://example.org/keys/odhfsd-sdfsdf-sdrqwe434-adasdasd-12311eawed",
                    "type": "Key",
                    "publicKeyPem": "<pem string of public key>",
                    "privateKeyPem": "<pem string of private ky>",
                }
            ]
        }
    )

    @override
    def __str__(self):
        return str(self.id)



"""
objects.py
This module defines the core object types for ActivityPub.
These objects are based on the ActivityStreams vocabulary.

For more information, see:
https://www.w3.org/TR/activitystreams-vocabulary/
"""

import enum
from typing import Any, ClassVar, cast, override

from pydantic import Field, HttpUrl, field_validator, model_validator
from pydantic.alias_generators import to_snake
from pydantic.config import ConfigDict

from phederation.models.proofs import DataIntegrityProof
from phederation.utils.base import (
    APDateTime,
    AccessType,
    ActivityPubBaseWithId,
    ObjectId,
    ObjectType,
)
from phederation.utils.exceptions import ValidationError as APValidationError
from phederation.utils.serialization import ActivityPubBase

from .links import APLink


class DocumentType(enum.Enum):
    """Document types, a subset of Object types that can be resolved to APDocument."""

    NOTE = "Note"
    ARTICLE = "Article"
    PAGE = "Page"
    DOCUMENT = "Document"
    IMAGE = "Image"
    VIDEO = "Video"
    AUDIO = "Audio"
    
    USERDATA = "UserData"

    @override
    def __str__(self):
        return str(self.value)


class APCollectionItem(ActivityPubBaseWithId):
    """Generic base class for storing ids of activity pub objects, together with access level, created date for paginatin, and object id."""

    type: str = ObjectType.COLLECTION_ITEM.value
    stored_id: ObjectId = Field(
        ...,
        description="Stored ObjectId. Different from the id property, because that one is the primary key and will lead to duplicates in the collections.",
    )
    collection_id: ObjectId = Field(..., description="Object identifier of the associated collection")
    visibility: str = Field(default=AccessType.PUBLIC.value, description="Access level (public, followers, private)")
    created_at: APDateTime | None = Field(default=None, description="Creation date")


def dereference(data: None | str | ObjectId | dict[str, Any] | ActivityPubBase, key: str) -> ObjectId | None:
    """Tries to get a URL/string property from the given key in the data dictionary.

    Args:
        data (dict[str, Any]): The data dictionary, possibly containing the given key.
        key (str): Key to search the URL for. The data[key] may be another json object, or an URL already.
    """
    if not data:
        return None
    data_dict: dict[str, Any]
    if isinstance(data, str) or isinstance(data, HttpUrl):
        data = {"id": str(data)}
    if isinstance(data, ActivityPubBase):
        snake_key = to_snake(key)
        data_dict = {key: getattr(data, snake_key, None)}
    else:
        data_dict = data
    if key not in data_dict:
        return None
    value = cast(str | dict[str, Any] | ActivityPubBaseWithId, data_dict[key])
    if isinstance(value, str) or isinstance(value, HttpUrl):
        return str(value)
    if isinstance(value, dict):
        value_id: str | None = value.get("id", None)
        if value_id is not None:
            return value_id
    if isinstance(value, ActivityPubBaseWithId):
        return getattr(value, "id", None)
    return None


def dereference_or_raise(data: None | str | ObjectId | dict[str, Any] | ActivityPubBase, key: str) -> ObjectId:
    value = dereference(data=data, key=key)
    if value is None:
        raise APValidationError(f"Key '{key}' not found in data {data}")
    return value


class APObject(ActivityPubBaseWithId):
    """Base class for all ActivityPub objects.

    According to ActivityStreams 2.0 Core Types: https://www.w3.org/TR/activitystreams-core/#object
    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#object
    Usage: https://www.w3.org/TR/activitypub/#object
    """

    type: str = Field(default="Object", description="The type of the object")
    attachment: None | list["str | APObject"] = Field(default=None, description="Files attached to the object")
    attributed_to: "str | None | APObject" = Field(default=None, description="Entity attributed to this object")
    content: str | None = Field(default=None, description="The content of the object")
    hash: str | None = Field(default=None, description="SHA256-Hash of the content. OPTIONAL")
    context: "None | str | APObject" = Field(default=None, description="Context of the object")
    name: str | None = Field(default=None, description="The name/title of the object")
    end_time: APDateTime | None = Field(default=None, description="End time of the object")
    generator: "None | str | APObject" = Field(default=None, description="Application that generated the object")
    icon: "None | str | APObject" = Field(default=None, description="Icon representing the object")
    image: "None | str | APObject" = Field(default=None, description="Image representing the object")
    in_reply_to: "None | str | APObject" = Field(default=None, description="Object this is in reply to")
    location: "None | str | APObject" = Field(default=None, description="Location associated with the object")
    preview: "None | str | APObject" = Field(default=None, description="Preview of the object")
    published: APDateTime | None = Field(default=None, description="Publication date")
    replies: "str | None | APObject" = Field(default=None, description="Replies to this object")
    start_time: APDateTime | None = Field(default=None, description="Start time of the object")
    summary: str | None = Field(default=None, description="Summary of the object")
    tag: None | list["str | APObject | APLink"] = Field(default=None, description="Tags associated with the object")
    updated: APDateTime | None = Field(default=None, description="Last update time")
    url: list[ObjectId] | None = Field(default=None, description="URL of the object")
    audience: list[ObjectId] | None = Field(default=None, description="Intended audience")
    to: list[ObjectId] | None = Field(default=None, description="Primary recipients")
    bto: list[ObjectId] | None = Field(default=None, description="Private primary recipients")
    cc: list[ObjectId] | None = Field(default=None, description="Secondary recipients")
    bcc: list[ObjectId] | None = Field(default=None, description="Private secondary recipients")
    media_type: str | None = Field(default=None, description="MIME type")
    visibility: str = Field(default="public", description="The visibility to actors other than the object's creator")
    proof: DataIntegrityProof | list[DataIntegrityProof] | None = Field(
        default=None, description="OPTIONAL Data integrity proof (or list of proofs) of the object."
    )

    likes: "None | str | APObject" = Field(default=None, description="Likes collection")
    shares: "None | str | APObject" = Field(default=None, description="Shares collection")

    @override
    def __str__(self):
        return str(self.id)

    def clear_recipients(self):
        self.to = None
        self.cc = None
        self.bto = None
        self.bcc = None
        self.audience = None

    async def get_recipients(self) -> list[ObjectId]:
        """Get the list of actors in the activity recipient fields (to,cc,bto,bcc,audience).
        Does *not* resolve the inboxes of these recipients, and also does *not* resolve collections (like followers etc.).

        Returns:
            list[ObjectId]: List of actor_id's for all recipients found.
        """
        recipients: list[ObjectId] = []
        recipients += self.to or []
        recipients += self.cc or []
        recipients += self.bto or []
        recipients += self.bcc or []
        recipients += self.audience or []

        recipients = list(set(recipients))
        return recipients

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "id": "https://example.org/id-of-the-object-unused",
                    "type": "Object",
                    "attributedTo": "https://example.org/id-of-an-actor",
                    "content": "The content of the object described in text.",
                    "summary": "A summary of the object.",
                }
            ]
        },
    )


class APEvent(APObject):
    """Event object as defined in ActivityStreams 2.0.
    An Event represents any kind of event that can occur, such as a concert, a meeting, or a conference.

    Validation:
        - end_time must be after start_time if both are provided
        - Inherits all validation from APObject

    https://www.w3.org/TR/activitystreams-vocabulary/#event
    """

    type: str = Field(default="Event", description="Indicates that this object represents an event")
    start_time: APDateTime | None = Field(default=None, description="The start time of the event")
    end_time: APDateTime | None = Field(default=None, description="The end time of the event")

    @model_validator(mode="after")
    def validate_end_time(self) -> "APEvent":
        """Validate that end_time is after start_time if both are provided."""
        if self.start_time and self.end_time:
            if self.end_time <= self.start_time:
                raise ValueError("end_time must be after start_time")
        return self


class APPlace(APObject):
    """Place object as defined in ActivityStreams 2.0.

    A Place represents a physical or logical location. It can be used to represent
    a geographic location, a venue, or any other kind of place.

    Validation:
        - latitude must be between -90 and 90
        - longitude must be between -180 and 180
        - units must be one of: 'cm', 'feet', 'inches', 'km', 'm', 'miles'
        - Inherits all validation from APObject

    https://www.w3.org/TR/activitystreams-vocabulary/#place
    """

    type: str = Field(default="Place", description="Indicates that this object represents a place")
    accuracy: float | None = Field(default=None, description="The accuracy of the coordinates (in meters)")
    altitude: float | None = Field(default=None, description="The altitude of the place (in meters)")
    latitude: float | None = Field(default=None, ge=-90, le=90, description="The latitude of the place (-90 to 90)")
    longitude: float | None = Field(default=None, ge=-180, le=180, description="The longitude of the place (-180 to 180)")
    radius: float | None = Field(default=None, description="The radius of uncertainty")
    units: str | None = Field(default=None, description="The units for the radius (e.g., 'cm', 'feet', 'km')")

    @field_validator("units")
    def validate_units(cls, v: str):
        valid_units = ["cm", "feet", "inches", "km", "m", "miles"]
        if v and v not in valid_units:
            raise ValueError(f"Invalid unit: {v}. Must be one of {', '.join(valid_units)}")
        return v

    @field_validator("latitude")
    def validate_latitude(cls, l: float):
        if l < -90 or l > 90:
            raise ValueError(f"Invalid latitude: {l}. Must be in the range of (-90, 90)")
        return l

    @field_validator("longitude")
    def validate_longitude(cls, l: float):
        if l < -180 or l > 180:
            raise ValueError(f"Invalid longitude: {l}. Must be in the range of (-180, 180)")
        return l


class APProfile(APObject):
    """
    Represents a Profile object.

    https://www.w3.org/TR/activitystreams-vocabulary/#profile
    """

    type: str = "Profile"
    describes: str | APObject


class APRelationship(APObject):
    """
    Represents a Relationship between objects.

    https://www.w3.org/TR/activitystreams-vocabulary/#relationship
    """

    type: str = "Relationship"
    subject: str | APObject
    object: str | APObject
    relationship: str


class APTombstone(APObject):
    """
    Represents a Tombstone (deleted object).

    https://www.w3.org/TR/activitystreams-vocabulary/#tombstone
    """

    type: str = "Tombstone"
    former_type: str
    deleted_at: APDateTime


class APDocument(APObject):
    """
    Represents a Document.

    https://www.w3.org/TR/activitystreams-vocabulary/#document
    """

    type: str = DocumentType.DOCUMENT.value
    hash: str | None = Field(
        default=None,
        description="SHA256-Hash of the document that is stored as media. If a file is stored and the document is describing it, the file hash is used here. OPTIONAL",
    )


class APArticle(APDocument):
    """
    Represents an Article.

    https://www.w3.org/TR/activitystreams-vocabulary/#article
    """

    type: str = DocumentType.ARTICLE.value


class APAudio(APDocument):
    """
    Represents an Audio object.

    https://www.w3.org/TR/activitystreams-vocabulary/#audio
    """

    type: str = DocumentType.AUDIO.value
    duration: str | None = None


class APImage(APDocument):
    """
    Represents an Image.

    https://www.w3.org/TR/activitystreams-vocabulary/#image
    """

    type: str = DocumentType.IMAGE.value
    width: int | None = Field(None, gt=0, description="The width of the image")
    height: int | None = Field(None, gt=0, description="The height of the image")


class APNote(APDocument):
    """
    Represents a Note.

    https://www.w3.org/TR/activitystreams-vocabulary/#note
    """

    type: str = DocumentType.NOTE.value


class APPage(APDocument):
    """
    Represents a Page.

    https://www.w3.org/TR/activitystreams-vocabulary/#page
    """

    type: str = DocumentType.PAGE.value


class APVideo(APDocument):
    """
    Represents a Video object.

    https://www.w3.org/TR/activitystreams-vocabulary/#video
    """

    type: str = DocumentType.VIDEO.value
    duration: str | None = None


class APUserData(APDocument):
    """
    Sending this object (by a normal user of an instance) will assemble their data into a file on the instance.
    The media url of the file can be accessed through the object of the maintenance activity.
    """

    type: str = DocumentType.USERDATA.value


"""
Base integration interfaces.
"""

from abc import ABC, abstractmethod

from fastapi import FastAPI

from phederation.api.base import BaseAPI
from phederation.federation.middleware import ActivityPubMiddleware
from phederation.federation.server import ActivityPubServer
from phederation.utils.exceptions import IntegrationError, catch_exceptions
from phederation.utils.settings import PhedSettings


class BaseIntegration(ABC):
    """Base integration interface."""

    app: FastAPI
    middleware: ActivityPubMiddleware
    server: ActivityPubServer

    def __init__(self, settings: PhedSettings):
        self.settings: PhedSettings = settings

    @catch_exceptions(IntegrationError, "Failed to initialize integration")
    @abstractmethod
    async def initialize(self) -> None:
        """Initialize integration."""
        pass

    @abstractmethod
    async def api(self, api_version: str | None = None) -> BaseAPI:
        """Create an API object depending on the API version specified in the request.
        If nothing is specified, use the latest API version.

        Args:
            api_version (str | None, optional): Version string with format "x.y.z", e.g. "1.0.0". Defaults to None.

        Returns:
            BaseAPI | None: The API version to return.
        """
        pass




from typing import override

from phederation.federation.tasks import BackgroundTask


class StorageBackgroundTask(BackgroundTask):

    @override
    async def on_startup(self):
        await super().on_startup()
        if self.server:
            await self.server.storage.initialize()

    @override
    async def on_shutdown(self):
        await super().on_shutdown()
        if self.server:
            await self.server.storage.close()



from typing import Annotated, ClassVar, override

from fastapi import Query
from pydantic import BaseModel
from pydantic.config import ConfigDict

from phederation.api.routes.base import BaseRoute
from phederation.models import RelativeLink
from phederation.utils.base import actor_id_from_username
from phederation.utils.exceptions import InvalidURLError, UserError


class WebfingerReponse(BaseModel):
    """Full webfinger response for a given actor."""

    subject: str
    aliases: list[str]
    links: list[RelativeLink]

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "subject": f"acct:username@example.org",
                    "aliases": ["username@example.org", "@username@example.org"],
                    "links": [
                        {
                            "rel": "self",
                            "type": "application/activity+json",
                            "href": "username@example.org",
                        },
                        {
                            "rel": "http://webfinger.net/rel/profile-page",
                            "type": "text/html",
                            "href": "username@example.org",
                        },
                    ],
                }
            ]
        }
    )


class WebfingerRoute(BaseRoute):

    @override
    def setup(self):

        @self.app.get("/.well-known/webfinger", tags=["instance"])
        async def get_webfinger(  # pyright: ignore[reportUnusedFunction]
            resource: Annotated[str, Query(description="Resource parameter. Must be of the form 'acct:username@instanceurl'.")],
        ) -> WebfingerReponse:
            """Handle WebFinger requests."""
            if not resource:
                raise InvalidURLError("Missing resource parameter")
            if not resource.startswith("acct:"):
                raise InvalidURLError("Invalid resource parameter, must be 'acct:username@instance_url'")

            username_domain = resource.replace("acct:", "").split("@")
            if len(username_domain) < 2:
                self.logger.warning(f"Invalid resource parameter: {username_domain}")
                raise InvalidURLError("Invalid resource parameter, must be 'acct:username@instance_url'")

            username, domain = username_domain[0], username_domain[1]

            port: str = ""
            if ":" in domain:
                port = f":{domain.split(':')[1]}"

            if self.api.settings.domain.url == domain or self.api.settings.domain.netloc == domain:
                user_id = actor_id_from_username(
                    base_url=self.api.settings.domain.base_url,
                    username=username,
                )
            else:
                self.logger.warning(f"Wrong resource: url_with_port={self.api.settings.domain.netloc}; while domain={domain}, port={port}.")
                raise InvalidURLError("The 'resource' parameter must be 'acct:username@instance_url'. Instance_url does not match current instance.")

            if not await self.server.storage.actor.read(id=user_id):
                raise UserError(f"Webfinger: User not found: {user_id}")

            response = WebfingerReponse(
                subject=f"acct:{username}@{self.api.settings.domain.url}{port}",
                aliases=[user_id],
                links=[
                    RelativeLink(rel="self", type="application/activity+json", href=user_id),
                    RelativeLink(
                        rel="http://webfinger.net/rel/profile-page",
                        type="text/html",
                        href=user_id,  # TODO: change this to a proper HTML page
                    ),
                ],
            )
            return response



from http import HTTPStatus
from typing import Annotated, override

from fastapi import Body, Depends, HTTPException, Security
from fastapi.responses import JSONResponse

from phederation.api.routes.base import BaseRoute
from phederation.models import (
    APUser,
)
from phederation.models.activities import APUpdate
from phederation.models.actors import APAccount, APActor
from phederation.security.authentication import UserInfo
from phederation.utils.exceptions import (
    ActivityPubException,
    AuthorizationError,
    UserError,
)


class AccountRoute(BaseRoute):

    @override
    def setup(self):

        @self.router.get(
            "/accounts/{username}",
            tags=["accounts"],
            response_model=APAccount,
            response_model_exclude_none=True,
        )
        async def get_account_route(  # pyright: ignore[reportUnusedFunction]
            actor: Annotated[APActor, Depends(self.middleware.authentication.verify_local_username)],
            user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])],
        ) -> APAccount:
            """Handle user account requests."""
            actor_for_authorization = user_info.actor_id if user_info else None

            if not actor.id or not actor_for_authorization or actor.id != actor_for_authorization:
                raise AuthorizationError(
                    f"Accounts can only be accessed by their associated actor. Actor '{actor_for_authorization}' is not the correct actor for account with id '{actor.id}'"
                )

            account = await self.api.server.storage.account.read(actor_for_authorization)
            if not account:
                raise UserError(f"Account for username '{actor.id}' not found")

            return account

        @self.router.put(
            "/accounts",
            tags=["users"],
            response_model_exclude_none=True,
            dependencies=[Security(self.middleware.authorization, scopes=["user"])],
        )
        async def update_account(  # pyright: ignore[reportUnusedFunction]
            user: Annotated[
                APUser,
                Body(description="The actor associated to the account."),
            ],
            user_account: Annotated[
                APAccount,
                Body(description="The information to update the account on the instance."),
            ],
        ):
            """Update the details of a user account."""
            activity = APUpdate(object=user_account, actor=user, summary="Update a given account.")
            try:
                actor_id = await self.server.update_account(activity=activity)
                return JSONResponse(
                    content={"Status": "OK", "location": actor_id, "detail": "User account updated successfully"},
                    status_code=HTTPStatus.OK,
                    media_type="application/activity+json",
                    headers={"Location": actor_id},
                )
            except ActivityPubException as e:
                self.logger.error(f"AccountRoute handler error in update_account: {e.message}")
                raise HTTPException(detail="User account could not be updated", status_code=HTTPStatus.BAD_REQUEST)



from typing import Any, cast, override

from redis.client import Redis

from phederation.cache.base import BaseCache


class RedisCache(BaseCache):
    def __init__(self, max_size: int = 100, ttl: int = 300, redis_host: str = "127.0.0.1", redis_port: int = 6379, password: None | str = None):
        self.max_size: int = max_size
        self.ttl: int = ttl
        self.r: Redis = Redis(host=redis_host, port=redis_port, password=password)

    @override
    def get(self, key: str) -> None | Any:
        return self.r.get(str(key))

    @override
    def set(self, key: str, value: Any | None):
        if value:
            _ = self.r.set(name=str(key), value=value)  # pyright: ignore[reportAny]

    @override
    def delete(self, key: str):
        _ = self.r.delete(str(key))

    @override
    def clear(self, prefix: str | None = None):
        if prefix is None:
            _ = self.r.flushall(asynchronous=False)  # pyright: ignore[reportUnknownMemberType]
        else:
            keys = cast(list[str], self.r.keys(f"{prefix}*"))  # pyright: ignore[reportUnknownMemberType]
            for key in keys:
                if key.startswith(prefix):
                    _ = self.r.delete(key)


    @override
    def close(self):
        self.r.close()



"""
tasks.py
Implements base classes for background tasks, mostly running periodically at certain intervals.
"""

from abc import ABC, ABCMeta, abstractmethod
import asyncio
from collections.abc import Callable, Coroutine
from contextlib import asynccontextmanager
from logging import Logger
from typing import Any

from fastapi import FastAPI

from phederation.utils.logging import configure_logger

from .server import ActivityPubServer


class BackgroundTask(ABC):
    server: ActivityPubServer

    def __init__(
        self,
    ):
        self.initialized: bool = False

    async def initialize(self, server: ActivityPubServer):
        """Called once the app is initialized, and the server is ready."""
        self.server = server
        self.initialized = True
        self.logger: Logger = configure_logger(__name__, prefix=server.settings.federation.logging_prefix + "_background_task")

    async def on_startup(self):
        """Called once on startup of the app. This can be replaced by implementing classes."""
        self.logger.info(f"Starting: {type(self).__name__}.")

    async def on_shutdown(self):
        """Called once on shutdown of the app. This can be replaced by implementing classes."""
        self.logger.info(f"Stopping: {type(self).__name__}.")

    def cancel(self):
        """Cancel the task, e.g. if no more updates should be done."""
        raise asyncio.CancelledError()


class PeriodicBackgroundTask(BackgroundTask, metaclass=ABCMeta):
    def __init__(self, update_period: int):
        super(PeriodicBackgroundTask, self).__init__()
        self.update_period: int = update_period

        if update_period <= 0:
            raise ValueError(f"Update period in PeriodicBackgroundTask ({type(self)}) must be positive (is {update_period})")

    @abstractmethod
    async def on_update(self):
        """Called every update tick."""


async def periodic_task(task: PeriodicBackgroundTask):
    while True:
        try:
            await task.on_update()
        except asyncio.CancelledError:
            break
        except Exception as e:
            task.logger.error(f"Exception in task {type(task).__name__}: {e}.")
        await asyncio.sleep(task.update_period)


def between_callback(some_callback: Callable[..., Coroutine[None, None, Any]]):
    def run():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        loop.run_until_complete(some_callback())
        loop.close()

    return run


def create_lifespan(logger_prefix: str, background_tasks: list[BackgroundTask]):
    """Create the periodic background tasks for the instance.
    Note that this will NOT work during testing, as the asyncio loop is different there.

    Returns:
        lifespan object: The lifespan async context generator for FastAPI.lifespan
    """

    @asynccontextmanager
    async def lifespan(app: FastAPI):  # pyright: ignore[reportUnusedParameter]
        lifespan_logger = configure_logger(__name__ + "_lifespan", prefix=logger_prefix)
        tasks_created: list[asyncio.Task[None]] = []
        if len(background_tasks) == 0:
            lifespan_logger.info("No background tasks configured.")
            yield
        else:
            lifespan_logger.info("Starting periodic background tasks...")
            for task in background_tasks:
                threaded_task = asyncio.to_thread(between_callback(task.on_startup))
                tasks_created.append(asyncio.create_task(threaded_task))
                if isinstance(task, PeriodicBackgroundTask):
                    task_created = asyncio.create_task(periodic_task(task))
                    tasks_created.append(task_created)
            yield
            # Run on shutdown (if required)
            lifespan_logger.info("Shutting down periodic background tasks...")
            for task in tasks_created:
                _ = task.cancel()
            for task in background_tasks:
                await task.on_shutdown()

    return lifespan



"""
Middleware between FastAPI and the activitypub implementation.
"""

from collections.abc import Awaitable, Callable
from datetime import datetime, timezone
from http import HTTPStatus
from logging import Logger
from typing import Annotated, Any

from fastapi import Depends, FastAPI
from fastapi import Security
from fastapi.requests import Request
from fastapi.responses import JSONResponse
from fastapi.responses import Response

from phederation.cache import BaseCache, WithCache
from phederation.cache.base import with_cache
from phederation.federation.ratelimit import RateLimiter
from phederation.federation.server import ActivityPubServer
from phederation.models import dereference
from phederation.models.activities import APActivity
from phederation.models.maintenance import APMaintenanceMode
from phederation.models.objects import dereference_or_raise
from phederation.security import Authentication
from phederation.security.authentication import UserInfo
from phederation.security.authorization import Authorization
from phederation.security.http_signatures import HTTPSignatureVerifier
from phederation.statistics.statistics import Statistics
from phederation.utils import ObjectId
from phederation.utils.base import AccessType, UrlType, serialize_datetime
from phederation.utils.exceptions import (
    AuthenticationError,
    MiddlewareError,
    SecurityError,
)
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings


class ActivityPubMiddleware(WithCache):
    """ActivityPub middleware."""

    def __init__(
        self,
        app: FastAPI,
        server: ActivityPubServer,
        statistics: Statistics,
        settings: PhedSettings,
        signature_verifier: HTTPSignatureVerifier | None,
        rate_limiter: RateLimiter,
        cache: BaseCache | None = None,
    ):
        self.settings: PhedSettings = settings
        self.server: ActivityPubServer = server
        self.statistics: Statistics = statistics
        self.signature_verifier: HTTPSignatureVerifier | None = signature_verifier
        self.rate_limiter: RateLimiter = rate_limiter
        self.app: FastAPI = app
        self.authentication: Authentication = Authentication(settings=settings, actor_manager=server.actor_manager)
        self.authorization: Authorization = Authorization(settings=settings, authentication=self.authentication)
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)
        self.cache: BaseCache | None = cache

        self.activity_from_token: Callable[[APActivity, UserInfo | None], Awaitable[APActivity]] | None = None
        self.activity_from_http: Callable[[APActivity, ObjectId | None], Awaitable[APActivity]] | None = None

    async def initialize(self):
        """Setup of all middleware routes."""
        self.logger.debug(f"Initializing middleware")

        @self.app.middleware("http")
        async def activitypub_middleware(request: Request, call_next: Callable[..., Awaitable[Response]]):  # pyright: ignore[reportUnusedFunction]
            """Processes all requests before they are handled internally."""
            request_url = str(request.url)
            self.logger.debug(f"Middleware processing request {request.method} at {request_url}")

            client_ip: str | None = request.headers.get("x-forwarded-for", None)
            if not client_ip and request.client:
                client_ip = f"{request.client.host}:{request.client.port}"
            else:
                client_ip = RateLimiter.EVERYBODY

            if self.settings.security.verify_ssl and request.url.scheme != "https":
                raise SecurityError(f"SSL verification failed in middleware for request url {request_url}")

            response_headers: dict[str, str] = {}
            if command := await self.check_maintenance_mode():
                if not self.settings.federation.admin_mode:
                    now = datetime.now(tz=timezone.utc)
                    if command.start_time:
                        response_headers["x-maintenance-start-time"] = serialize_datetime(command.start_time)
                    if command.end_time:
                        response_headers["x-maintenance-end-time"] = serialize_datetime(command.end_time)
                    if command.id:
                        response_headers["x-maintenance-id"] = command.id
                    if (not command.start_time or command.start_time <= now) and (not command.end_time or now <= command.end_time):
                        response_dict = {"status": "maintenance", "detail": "In maintenance mode."}
                        for key, value in response_headers.items():
                            response_dict[key] = value
                        return JSONResponse(status_code=HTTPStatus.METHOD_NOT_ALLOWED, content=response_dict, headers=response_headers)

            # test if the request includes a valid HTTP signature; which then can enable higher rate limits
            try:
                rate_limit_domain = await _verify_http_signature(request=request, verified_actor_key_id=None)
            except AuthenticationError:
                # if not authenticated, rate limit on ip
                rate_limit_domain = client_ip

            if self.rate_limiter.check_rate_limit(domain=rate_limit_domain):
                response_headers = self.rate_limiter.update_headers(domain=rate_limit_domain, headers=response_headers)

                response = await call_next(request)
                response = await self.process_response(response, response_headers=response_headers)
            else:
                response = self.rate_limiter.ratelimit_response(rate_limit_domain)

            response = await self.statistics.log(request=request, response=response)
            return response

        async def _verify_http_signature(
            request: Request, verified_actor_key_id: Annotated[ObjectId | None, Depends(self.get_key_from_activity)]
        ) -> ObjectId:
            """
            Verify signature of incoming HTTP request.

            Args:
                request: Incoming Request to verify HTTP signature

            Returns:
                ObjectID: key id used to sign the request.

            Raises:
                MiddlewareError: If the middleware is not set up properly.
                AuthenticationError: If the HTTP request is signed improperly (or not signed at all).
            """
            # Verify HTTP signature, including body
            if self.signature_verifier:
                # TODO: does this pull an entire file if a media file is sent? This should not be the case,
                # file hashes are computed in media.py and iteratively.
                body_serialized = (await request.body()).decode(encoding="utf-8")
                key_id = await self.signature_verifier.verify_signature(
                    headers=dict(request.headers),
                    body_serialized=body_serialized,
                    verified_key_id=verified_actor_key_id,
                    method=request.method,
                    path=request.url.path,
                )
                if not key_id:
                    raise AuthenticationError(f"Signature could not be verified. Method={request.method}, path={request.url.path}")
            else:
                raise MiddlewareError(
                    f"Signature could not be verified: Middleware does not have a signature verifier object, will just stop all requests."
                )

            return key_id

        async def _activity_from_token(
            activity: APActivity,
            user_info: Annotated[UserInfo | None, Security(self.authentication, scopes=["user"])],
        ) -> APActivity:
            if user_info and user_info.actor_id and user_info.actor_id == dereference(activity, "actor"):
                return activity
            raise AuthenticationError(f"Activity {activity.type} actor could not be authenticated from token")

        async def _activity_from_http(
            activity: APActivity,
            key_id: Annotated[ObjectId | None, Security(_verify_http_signature, scopes=["user"])],
        ) -> APActivity:
            actor_id = dereference_or_raise(activity, "actor")
            if key_id:
                # dereference public key from activity actor, check if they are the same as the http key id
                actor_key_id = await self.server.key_manager.get_key_id_from_actor(actor_id)
                if actor_key_id and actor_key_id == key_id:
                    return activity
                else:
                    raise AuthenticationError(f"Activity {activity.type} actor could not be authenticated from HTTP signature, keys do not match")
            raise AuthenticationError(f"Activity {activity.type} actor could not be authenticated from HTTP signature, key_id is None")

        self.activity_from_token = _activity_from_token
        self.activity_from_http = _activity_from_http

    @with_cache(prefix=__name__)
    async def check_maintenance_mode(self) -> APMaintenanceMode | None:
        """Check if the maintenance mode has been scheduled by an admin.

        Returns:
            APMaintenanceMode | None: Either APMaintenanceMode command if maintenance mode was scheduled, or None.
        """
        maintenance_commands_id = self.server.instance.maintenance_commands
        command_ids = await self.server.collections.get_collection_items(collection_id=maintenance_commands_id, access=AccessType.PRIVATE)
        for id in command_ids:
            self.logger.critical(f"Processing maintenance id in commands: {id}")
            maintenance_command = await self.server.resolver.try_resolve_from_storage(url=id, url_type=UrlType.Objects)
            self.logger.critical(f"Resolved command: {maintenance_command}")
            if isinstance(maintenance_command, APMaintenanceMode):
                now = datetime.now(timezone.utc)
                if not maintenance_command.end_time or maintenance_command.end_time > now:
                    return maintenance_command
        else:
            return None

    async def get_key_from_activity(self, activity: APActivity) -> ObjectId | None:
        """Get the id of the public key of the actor of the given activity.

        The activity.actor does not need to be local to the instance.

        Args:
            activity (APActivity): An activity with an actor.

        Returns:
            ObjectId | None: The id (an url) of the public key, or None.
        """
        actor_id = dereference(activity, "actor")
        if not actor_id:
            return None
        key_id = await self.server.key_manager.get_key_id_from_actor(actor_id=actor_id)
        return key_id

    async def process_response(self, response: Response, response_headers: dict[str, Any] | None = None) -> Response:
        """
        Process outgoing response.

        Args:
            response: Response from the call.

        Returns:
           Response with processed headers.
        """

        # Add standard headers for Json responses
        headers_lower = {key.lower(): response.headers.get(key) for key in response.headers.keys()}
        content_type = response.headers["content-type"] if "content-type" in headers_lower else None
        content_type = response.headers["accept"] if "accept" in headers_lower else content_type
        if content_type and "json" in content_type:
            response.headers.update({"Vary": "Accept, Accept-Encoding", "Cache-Control": "max-age=0, private, must-revalidate"})
            if response_headers:
                response.headers.update(response_headers)

        return response



"""
worker_encryption.py
"""

from collections.abc import Iterable
from logging import Logger
from typing import override

from cryptography.fernet import Fernet
from temporalio.api.common.v1 import Payload
from temporalio.converter import PayloadCodec

from phederation.utils.exceptions import SecurityError
from phederation.utils.logging import configure_logger


class EncryptionCodec(PayloadCodec):
    """Handles compression and encryption of payloads to temporalio and back.

    Note that only the temporalio backend does not see the private keys,
    methods starting activities/workflows and workers handling them all run on the local instance and have access to everything.

    Encryption is done using the given key_id (typically the instance actor key), and RSA encryption with PKCS1v15 padding.

    See: https://docs.temporal.io/develop/python/converters-and-encryption
    """

    def __init__(self, fernet_key: bytes, logging_prefix: str) -> None:
        super().__init__()
        self.key: bytes = fernet_key
        self.logger: Logger = configure_logger(__name__, logging_prefix)

    @override
    async def encode(self, payloads: Iterable[Payload]) -> list[Payload]:
        return [
            Payload(
                metadata={
                    "encoding": b"binary/encrypted",
                    "encryption-key-id": b"fernet",
                },
                data=await self.encrypt(data=p.SerializeToString()),
            )
            for p in payloads
        ]

    @override
    async def decode(self, payloads: Iterable[Payload]) -> list[Payload]:
        ret: list[Payload] = []
        for p in payloads:  # Ignore ones w/out our expected encoding
            if p.metadata.get("encoding", b"").decode() != "binary/encrypted":
                ret.append(p)
                continue
            # Confirm our key ID is the same
            key_id = p.metadata.get("encryption-key-id", b"").decode(encoding="utf-8")
            if key_id != "fernet":
                raise SecurityError(f"Unrecognized key ID {key_id}. Current key ID is fernet.")
            # Decrypt and append
            ret.append(Payload.FromString(await self.decrypt(data=p.data)))
        return ret

    async def encrypt(self, data: bytes) -> bytes:
        try:
            return Fernet(self.key).encrypt(data)
        except Exception as e:
            self.logger.error(f"Error in encrypting for temporalio: {e}")
            raise SecurityError("Failed to encrypt message to temporalio")

    async def decrypt(self, data: bytes) -> bytes:
        try:
            return Fernet(self.key).decrypt(data)
        except Exception as e:
            self.logger.error(f"Error in decrypting from temporalio: {e}")
            raise SecurityError("Failed to decrypt message from temporalio")



"""
JSON serializer for ActivityPub objects.
"""

import json
from typing import Any, ClassVar

from pydantic import BaseModel
from pydantic.alias_generators import to_camel
from pydantic.config import ConfigDict


def to_json_string(data: dict[str, Any]) -> str:
    """
    Convert dictionary to JSON string with standardized formatting.

    Args:
        data: Dictionary to convert

    Returns:
        JSON string with consistent formatting
    """
    return json.dumps(data, sort_keys=True, ensure_ascii=True, separators=(",", ":"))


class ActivityPubBase(BaseModel):
    """Base class for all ActivityPub objects."""

    def serialize(self, include_context: bool = True) -> dict[str, Any]:
        """Serialize the AP object to a dict.

        Args:
            include_context (bool, optional): If the JSON-LD context for ActivityPub should be included in the dict. Defaults to True.

        Returns:
            dict[str, Any]: JSON-LD representation of the object.
        """
        processed_data = self.model_dump(by_alias=True, exclude_none=True)

        # Add context if needed
        if include_context:
            context = ["https://www.w3.org/ns/activitystreams", "https://w3id.org/security/v1"]
            if "proof" in processed_data and isinstance(processed_data["proof"], dict):
                context.append("https://w3id.org/security/data-integrity/v2")
            processed_data["@context"] = context

        return processed_data

    @classmethod
    def deserialize(cls, data: str | dict[str, Any]) -> "ActivityPubBase":
        """Deserialize dict to an ActivityPub object.
        This is removing the json-ld context before deserialization.
        """
        if isinstance(data, dict):
            data.pop("@context", None)
        obj = cls.model_validate(data)
        return obj

    model_config: ClassVar[ConfigDict] = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        extra="allow",
        arbitrary_types_allowed=True,
        loc_by_alias=True,
    )



"""
security/http_signatures.py
Enhanced HTTP Signatures implementation.

Implements HTTP Signatures (draft-cavage-http-signatures) of request bodies.

See https://swicg.github.io/activitypub-http-signature/#how-to-verify-a-signature
"""

import base64
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
import hashlib
from logging import Logger
from urllib.parse import urlparse

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding

from phederation.utils import ObjectId
from phederation.utils.exceptions import (
    SignatureError,
    catch_exceptions,
)
from phederation.utils.logging import configure_logger

from .key_management import KeyManager, KeyPair


@dataclass
class HeaderSignature:
    algorithm: str
    key_id: str
    signed_headers: list[str]
    signature: bytes

    def build(self) -> str:
        return (
            f'keyId="{self.key_id}",'
            f"algorithm={self.algorithm},"
            f'headers="{" ".join(self.signed_headers)}",'
            f'signature="{base64.b64encode(self.signature).decode(encoding='utf-8')}"'
        )

    @catch_exceptions(SignatureError, "Invalid header signature")
    @staticmethod
    async def parse(header: str) -> "HeaderSignature":
        """Parse HTTP signature header."""
        parts: dict[str, str] = {}
        for part in header.split(","):
            if "=" not in part:
                continue
            key, value = part.split("=", 1)
            parts[key.strip()] = value.strip('"')

        required = ["keyId", "algorithm", "headers", "signature"]
        if not all(k in parts for k in required):
            raise SignatureError(f"Missing required signature parameters (is {header}, should be {required})")

        signature_bytes = base64.b64decode(parts["signature"])
        return HeaderSignature(
            algorithm=parts["algorithm"], key_id=parts["keyId"], signed_headers=parts["headers"].split(), signature=signature_bytes
        )


class HTTPSignatureVerifier:
    """HTTP signature verification.

    See https://swicg.github.io/activitypub-http-signature/#how-to-verify-a-signature
    """

    def __init__(self, key_manager: KeyManager):
        """Initialize signature verifier."""
        self.key_manager: KeyManager = key_manager
        self.logger: Logger = configure_logger(__name__, prefix=key_manager.storage.settings.federation.logging_prefix)

    @catch_exceptions(SignatureError, "Could not verify request signature")
    async def verify_signature(
        self, headers: dict[str, str], method: str, path: str, verified_key_id: None | str, body_serialized: None | str
    ) -> ObjectId | None:
        """Verify HTTP signature on request."""
        headers = {key.lower(): headers[key] for key in headers.keys()}

        # then verify signature
        if "signature" not in headers:
            self.logger.debug("No 'signature' field found in headers")
            return None

        # verify TTL first
        if "date" not in headers:
            self.logger.warning("No 'date' field found in headers")
            return None

        if not self._verify_date(headers["date"]):
            self.logger.warning(f"Http signature expired, TTL is {self.key_manager.settings.security.signature_ttl} seconds")
            return None

        sig_parts = await HeaderSignature.parse(header=headers["signature"])

        # Verify signing algorithm
        allowed_algorithms = self.key_manager.settings.security.allowed_algorithms
        if sig_parts.algorithm not in allowed_algorithms:
            self.logger.warning(f"Algorithm '{sig_parts.algorithm}' not in allowed algorithms ({allowed_algorithms})")
            return None

        # Get signing key
        if verified_key_id and verified_key_id != sig_parts.key_id:
            self.logger.warning(f"keyId in signature is not equal to verified_key_id ('{sig_parts.key_id}' is not '{verified_key_id}')")
            return None

        # If body is provided and digest header exists, verify the digest
        if self.key_manager.settings.security.require_body_digest:
            if body_serialized is None or "digest" not in headers:
                self.logger.warning("Body digest verification failed")
                return None
            expected_digest = self._generate_digest(body_serialized=body_serialized)
            if headers["digest"] != expected_digest:
                self.logger.warning("Body digest verification failed")
                return None

        signing_string = self._build_signing_string(method=method, path=path, signed_headers=sig_parts.signed_headers, headers=headers)

        # Get public key, possibly resolving it from the network
        public_key = await self.key_manager.get_public_key(key_id=sig_parts.key_id)
        if not public_key:
            self.logger.error(f"Failed to get public key for {sig_parts.key_id}")
            return None

        # Verify signature
        try:
            public_key.verify(sig_parts.signature, signing_string.encode(encoding="utf-8"), padding.PKCS1v15(), hashes.SHA256())
            return sig_parts.key_id
        except InvalidSignature as e:
            self.logger.warning(f"Invalid signature, error: '{e}', sig_parts '{sig_parts}")
            return None

    @catch_exceptions(SignatureError, "sign request failed")
    async def sign_request(
        self, method: str, path: str, headers: dict[str, str], key_id: ObjectId, body_serialized: None | str = None
    ) -> dict[str, str]:
        """Sign HTTP request."""
        # make sure to remove the host
        path = urlparse(path).path

        request_headers = headers.copy()

        # Add date if not present
        if "date" not in request_headers:
            now = datetime.now(timezone.utc)
            request_headers["date"] = now.strftime("%a, %d %b %Y %H:%M:%S GMT")

        # Calculate digest first
        if body_serialized is not None:
            digest = self._generate_digest(body_serialized)
            request_headers["digest"] = digest
            self.logger.debug(f"Added digest header: {digest}")

        # Headers to sign (in specific order)
        headers_to_sign = ["(request-target)", "host", "date", "digest"]

        # Build signing string
        signing_string = self._build_signing_string(method, path, request_headers, headers_to_sign)

        self.logger.debug(f"Headers being signed: {headers_to_sign}. Signing string: {signing_string}")

        # Get private key
        if self.key_manager:
            active_key: KeyPair = await self.key_manager.get_RSA_key_pair(key_id=key_id)
        else:
            raise SignatureError("No key manager available")
        if not active_key.private_key:
            raise SignatureError(f"HTTPSignatureVerifier: Key id '{key_id}' does not have a local private key.")

        # Sign
        signature = active_key.private_key.sign(signing_string.encode(encoding="utf-8"), padding.PKCS1v15(), hashes.SHA256())

        # Build signature header
        signature_header = HeaderSignature(
            key_id=active_key.key_id, algorithm='"rsa-sha256"', signed_headers=headers_to_sign, signature=signature
        ).build()

        # Return headers with both Digest and Signature
        signed_headers = {**request_headers, "Signature": signature_header}

        self.logger.debug(f"Final headers: {signed_headers}")
        return signed_headers

    def _verify_date(self, date_header: None | str) -> bool:
        """
        Verify request date with clock skew handling.

        Allows 5 minutes of clock skew in either direction.
        """
        if not date_header:
            return False

        try:
            request_time = datetime.strptime(date_header, "%a, %d %b %Y %H:%M:%S GMT").replace(tzinfo=None)

            # Use test time if set, otherwise use current time
            now = datetime.now(timezone.utc)
            now = now.replace(tzinfo=None)

            skew = timedelta(minutes=5)
            earliest = now - skew
            latest = now + skew

            self.logger.debug(f"Date verification: request_time={request_time}, now={now}, earliest={earliest}, latest={latest}")

            return earliest <= request_time <= latest

        except Exception as e:
            self.logger.debug(f"Date verification failed: {e}")
            return False

    def _build_signing_string(self, method: str, path: str, headers: dict[str, str], signed_headers: list[str]) -> str:
        """Build string to sign."""
        lines: list[str] = []
        # Convert headers to case-insensitive dictionary
        headers_lower = {k.lower(): v for k, v in headers.items()}

        for header in signed_headers:
            if header == "(request-target)":
                lines.append(f"(request-target): {method.lower()} {path}")
            else:
                header_lower = header.lower()
                if header_lower not in headers_lower:
                    self.logger.error(f"Missing required header: {header}. Available headers: {list(headers_lower.keys())}")
                    raise SignatureError(f"Missing required header: {header}")
                lines.append(f"{header_lower}: {headers_lower[header_lower]}")

        signing_string = "\n".join(lines)
        self.logger.debug(f"Signing string: {signing_string}")
        return signing_string

    def _generate_digest(self, body_serialized: str) -> str:
        """
        Generate digest header value for request body.

        Args:
            body_serialized: Request body string (already serialized)

        Returns:
            Digest header value
        """
        body_bytes = body_serialized.encode("utf-8")

        # Calculate SHA-256 digest
        digest = hashlib.sha256(body_bytes).digest()
        digest_b64 = base64.b64encode(digest).decode("ascii")

        return f"SHA-256={digest_b64}"



"""
Handler for Follow activities.
"""

from datetime import datetime
from typing import override

from phederation.models import APAccept, ValidCollection, dereference
from phederation.models.activities import APActivity, APFollow
from phederation.utils.base import AccessType, collection_id_from_name
from phederation.utils.exceptions import HandlerError, ValidationError

from .base import ActivityHandler


class FollowHandler(ActivityHandler):
    """Handles Follow activities."""

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Follow activity."""
        if activity.type != "Follow":
            raise ValidationError("Invalid activity type")

        if not activity.actor:
            raise ValidationError("Follow must have an actor")

        # Check that the object of the activity exists.
        if not activity.object:
            raise ValidationError("Follow must have an object")

        # make sure we are not following ourselves
        if dereference(data=activity, key="actor") == dereference(data=activity, key="object"):
            raise ValidationError("An actor cannot follow itself.")

        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        actor_id = dereference(activity, key="actor")
        target_id = dereference(activity, key="object")

        # Resolve target actor
        target_actor = await self.resolver.resolve_actor(target_id)
        if not target_actor or not target_id or not actor_id:
            raise HandlerError(f"Could not resolve target actor: {target_id}")

        # Check if already following (just if it was registered previously, not if it was actually accepted)
        # TODO: this does not work without proper storage session management; the await of is_following is a problem
        if not await self.storage.is_following(actor_id, target_id, check_accepted=False):
            self.logger.debug(f"FollowHandler: storage does not have the follow activity '{activity.id}' already, storing it now.")
            # Store follow request
            follow_id = await self.storage.follow.create(
                data=APFollow(id=activity.id, accepted=False, actor=actor_id, object=target_id), raise_if_exists=False
            )
            collection_follow = collection_id_from_name(id=actor_id, name=ValidCollection.Following.value)
            access = AccessType.from_visibility(activity.visibility)
            _ = await self.collections.add_to_collection(collection_id=collection_follow, items=follow_id, access=access, add_only_once=True)

        # Inform the target actor. This should trigger an "ACCEPT" from their side (if accepted).
        send_follow_result = await self.delivery.deliver_to_actor_inbox(activity=activity, actor_id=target_id)
        if send_follow_result.failed:
            self.logger.warning(f"Failing to send the Follow activity to the actor to follow: {send_follow_result}.")
        return activity

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        follow_activity_id = activity.id
        follower_id = dereference(activity.actor, "id")
        following_id = dereference(activity, key="object")

        if not follow_activity_id or not following_id or not follower_id:
            raise HandlerError("Could not resolve follow activity id, following_id, or follower_id")

        # Resolve target actor
        # Check that the object of the activity is an actor that can be followed. Actors have at least an inbox and an outbox property.
        target_actor = await self.resolver.resolve_actor(following_id)
        if not target_actor or not following_id:
            raise HandlerError(f"Could not resolve target actor: {following_id}")

        self.logger.debug(f"FollowHandler: starting to have follower={follower_id} to following={following_id}...")

        # Check if already following (just if it was registered previously, not if it was actually accepted)
        # also check that an outstanding Follow request has not already been made. It may be useful to store outstanding Follow requests in an easy-to-find way in your storage.
        if await self.storage.is_following(follower=follower_id, following=following_id, check_accepted=True):
            return None

        # Check that the object of the activity is an object on the local server.
        # Otherwise, the Follow activity can be safely treated like any other activity with no side effects.
        if not target_actor.local:
            return None

        # If these checks work, there are two paths to follow:
        if not target_actor.autoaccept:
            # Path 1: If the recipient does require manually confirming follows, the server should allow the actor to review the follow and either accept or reject. A typical implementation shows new follow requests in a separate section of the UI. Another option is including the Follow activity in the actor's inbox for review.
            self.logger.debug(f"FollowHandler: following actor requires manual accept, waiting...")
            if not await self.storage.is_following(follower=follower_id, following=following_id, check_accepted=False):
                follow_id = await self.storage.follow.create(
                    APFollow(
                        id=activity.id,
                        actor=follower_id,
                        object=following_id,
                        accepted=False,
                    )
                )
                # do not add to the followers collection here, actor has not accepted yet.
                return follow_id
        else:
            # Path 2: If the recipient does not require manually confirming follows, then you can simply create the connection and send an Accept activity in response.
            self.logger.debug(f"FollowHandler: following actor does autoaccept, storing follow for activity '{activity.id}'.")
            if not activity.id:
                raise HandlerError(f"FollowHandler: activity.id is None")
            _ = await self.storage.follow.upsert(
                id=activity.id,
                data=APFollow(
                    id=activity.id, actor=follower_id, object=following_id, accepted=True, accepted_at=datetime.now(), visibility=activity.visibility
                ),
            )
            # send Accept back to follower
            accept = APAccept(
                actor=following_id,
                object=follow_activity_id,
                to=[follower_id],  # this is necessary here, otherwise the follower actor will not be notified
                visibility=activity.visibility,
            )

            # To create the connection, there are two steps:
            #     1. Add the actor of the activity to the object's followers collection (done here, on this side).
            #     2. Add the object of the activity to the actor's following collection (done on the follower side).
            collection_follow = collection_id_from_name(id=following_id, name=ValidCollection.Followers.value)
            access = AccessType.from_visibility(activity.visibility)
            _ = await self.collections.add_to_collection(collection_id=collection_follow, items=follow_activity_id, access=access, add_only_once=True)

            # deliver the accept to the outbox of the actor that is being followed (to register the Accept and then send it off)
            following_dict = await self.resolver.resolve_actor(actor_id=following_id)
            following_outbox = dereference(data=following_dict, key=ValidCollection.Outbox.value)

            if not following_outbox:
                raise HandlerError(f"FollowHandler: outbox of actor is None; actor_dict={following_outbox}")

            # deliver the accept to the outbox of the actor being followed.
            # Note that this is NOT using token based authentication, but directly goes into the outbox of the actor on the local instance.
            result = await self.delivery.deliver_to_local_actor(activity=accept, collection=ValidCollection.Outbox, actor_id=following_id)
            if result.failed:
                self.logger.warning(f"FollowHandler: Accept activity delivery failed: {result.failed}, error message: {result.error_message}")
            return follow_activity_id



from typing import Annotated, override

from fastapi import Security

from phederation.api.routes.base import BaseRoute
from phederation.models.collections import APOrderedCollection
from phederation.models.objects import APObject, dereference
from phederation.security.authentication import UserInfo
from phederation.utils.base import (
    AccessType,
    ObjectType,
    UrlType,
    assemble_id_url,
    collection_id_from_name,
)
from phederation.utils.exceptions import AuthorizationError, ValidationError


class ObjectRoute(BaseRoute):

    @override
    def setup(self):

        # "response_model=None" is important here, because this allows to return fields of classes that inherit from APObject.
        # If it is not set, FastAPI will *only* return fields from APObject, and none from inheriting classes (like APDocument).
        @self.app.get("/objects/{object_primary}", tags=["objects"], response_model_exclude_none=True, response_model=None)
        async def get_object(  # pyright: ignore[reportUnusedFunction]
            object_primary: str, user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])] = None
        ) -> APObject:
            """Return object with given id, if accessible."""
            object_id = assemble_id_url(
                base_url=self.api.settings.domain.base_url,
                primary=object_primary,
                type=UrlType.Objects,
            )
            obj = await self.server.storage.object.read(id=object_id)
            if not obj:
                raise ValidationError("Object does not exist locally")

            # TODO: security: unit test for this visibility check
            actor_id = dereference(obj, "attributed_to")
            actor_for_authorization = user_info.actor_id if user_info else None
            access: AccessType = await self.server.determine_access(actor_id=actor_id, actor_for_authorization=actor_for_authorization)
            if not AccessType.from_visibility(obj.visibility).is_accessible(access):
                raise AuthorizationError(f"Object {object_id} cannot be accessed by actor {actor_for_authorization}, visibility is {obj.visibility}")

            # Add DataIntegrityProof here if required by settings (unless it is a tombstone)
            if self.server.settings.media.proof_object_hash and obj.type != ObjectType.TOMBSTONE.value:
                if obj.hash is None:
                    raise ValidationError(f"Cannot add proof to object {object_id}: object.hash is None")
                actor_id = dereference(obj, "attributed_to")
                if actor_id is None:
                    raise ValidationError(f"Cannot resolve object.attributed_to")
                private_key_id = await self.server.key_manager.get_key_id_from_actor(actor_id)
                obj.proof = await self.server.key_manager.proof_string(string_to_sign=obj.hash, private_key_id=private_key_id)

            # add information about the collections of the object
            obj.shares = f"{object_id}/shares"
            obj.likes = f"{object_id}/likes"
            return obj

        @self.app.get(
            "/objects/{object_id}/shares",
            response_model_exclude_none=True,
            tags=["objects"],
        )
        async def get_object_shares(  # pyright: ignore[reportUnusedFunction]
            object_id: str,
            page: int | None = None,
            user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])] = None,
        ) -> APOrderedCollection:
            """Return object shares list."""
            object_id = assemble_id_url(
                base_url=self.api.settings.domain.base_url,
                primary=object_id,
                type=UrlType.Objects,
            )
            obj = await self.server.storage.object.read(id=object_id)
            if not obj:
                raise ValidationError("Object does not exist locally.")
            actor_id = dereference(obj, "attributed_to")
            collection_id = collection_id_from_name(id=object_id, name="shares", page=None)
            return await self.server.handle_pagination(
                collection_id=collection_id, actor_id=actor_id, actor_for_authorization=user_info.actor_id if user_info else None, page=page
            )

        @self.app.get(
            "/objects/{object_id}/likes",
            response_model_exclude_none=True,
            tags=["objects"],
        )
        async def get_object_likes(  # pyright: ignore[reportUnusedFunction]
            object_id: str,
            page: int | None = None,
            user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])] = None,
        ) -> APOrderedCollection:
            """Return object likes list."""
            object_id = assemble_id_url(
                base_url=self.api.settings.domain.base_url,
                primary=object_id,
                type=UrlType.Objects,
            )
            obj = await self.server.storage.object.read(id=object_id)
            if not obj:
                raise ValidationError("Object does not exist locally.")

            collection_id = collection_id_from_name(id=object_id, name="likes", page=None)
            actor_id = dereference(obj, "attributed_to")
            return await self.server.handle_pagination(
                collection_id=collection_id, actor_id=actor_id, actor_for_authorization=user_info.actor_id if user_info else None, page=page
            )



"""
version.py
Contains the version string of the current instance.
Useful for UserAgent, and database migration.
"""

import os
from typing import cast

import toml

pyproject_toml_path = "pyproject.toml"
if os.path.exists(pyproject_toml_path):
    with open(pyproject_toml_path, "r") as f:
        pyproject_toml = toml.load(f)
        version_string = cast(str, pyproject_toml["project"]["version"])
else:
    version_string = "0.3"


PHEDERATION_SOFTWARE_NAME = "phederation"
PHEDERATION_VERSION = version_string
PHEDERATION_USERAGENT = f"{PHEDERATION_SOFTWARE_NAME}/{PHEDERATION_VERSION}"

PHEDERATION_HEADERS = {
    "Content-Type": 'application/ld+json; profile="https://www.w3.org/ns/activitystreams"',
    "User-Agent": PHEDERATION_USERAGENT,
    "Accept": "application/activity+json"
}


"""
Update activity handler.
"""

from datetime import datetime, timezone
from typing import override
from urllib.parse import urlparse, urlunparse

from phederation.handlers.delete import DeleteHandler
from phederation.models import APLink, APObject, DataIntegrityProof
from phederation.models.activities import APActivity
from phederation.models.actors import APActor
from phederation.models.collections import APCollection
from phederation.models.keys import APPrivateKey
from phederation.models.objects import dereference, dereference_or_raise
from phederation.utils import ActivityType, NodeInfo, ObjectId, UrlType
from phederation.utils.base import AccessType, CollectionType, assemble_id_url
from phederation.utils.exceptions import (
    AuthorizationError,
    HandlerError,
    ResolverError,
    ValidationError,
)

from .base import ActivityHandler


class UpdateHandler(ActivityHandler):
    """Handle Update activities."""

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Update activity."""
        if activity.type != ActivityType.UPDATE.value:
            raise ValidationError("Invalid activity type")

        object_id = dereference(activity, "object")
        if not object_id:
            raise ValidationError("Missing object")

        actor_id = dereference(activity, "actor")
        if not actor_id:
            raise ValidationError("Missing actor")

        # check if the object is a collection instead of an object
        if await DeleteHandler.check_is_collection(self, actor_id=actor_id, object_id=object_id):
            # validate that the new collection name still is a collection of the actor
            collections_of_actor = assemble_id_url(type=UrlType.Collections, base_url=actor_id)
            if not object_id.startswith(collections_of_actor):
                raise ValidationError(f"New collection id {object_id} is not in the collections of actor {actor_id}")
            try:
                obj = await self.resolver.resolve_object(object_id)
                if not isinstance(obj, APCollection):
                    raise ValidationError("Object of APUpdate for collection is not of type APCollection")
            except ResolverError as e:
                raise ValidationError(e.message)
        else:
            # Validate object exists
            obj = activity.object
            if isinstance(obj, dict):
                obj = APObject.deserialize(obj)
            if not isinstance(obj, APObject):
                raise ValidationError("Invalid object format")
            object_id = dereference_or_raise(obj, "id")
            object_type = dereference_or_raise(obj, "type")

            url_type = UrlType.from_object_type(object_type)
            existing = await self.resolver.try_resolve_from_storage(url=object_id, url_type=url_type)
            if not existing:
                raise ValidationError(f"Object not found: {object_id}")
            if isinstance(existing, NodeInfo):
                raise ValidationError(f"Object type of object in storage is wrong: {type(existing).__name__}")

            if isinstance(existing, DataIntegrityProof):
                raise ValidationError(f"Object of type DataIntegrityProof cannot be updated")

            # Validate actor has permission
            if existing.attributed_to != activity.actor:
                raise AuthorizationError("Not authorized to update object")

        # set to the deserialized and validated object
        activity.object = obj

        return activity

    async def update_collection(self, actor_id: ObjectId, collection_id: ObjectId, collection_id_new: ObjectId):
        """Update the name of a collection."""
        self.logger.debug(f"Updating collection name from {collection_id} to {collection_id_new}")
        # TODO: use storage to rename the collection, dont remove/add
        collection_items_public = await self.collections.get_collection_items(collection_id=collection_id, access=AccessType.PUBLIC)
        collection_items_follower = await self.collections.get_collection_items(collection_id=collection_id, access=AccessType.FOLLOWER)
        collection_items_private = await self.collections.get_collection_items(collection_id=collection_id, access=AccessType.PRIVATE)
        for item in collection_items_public:
            collection_items_private.remove(item)
            collection_items_follower.remove(item)
        for item in collection_items_follower:
            collection_items_private.remove(item)
        await self.collections.delete_collection(collection_id=collection_id)

        # also remove the old collection name from the actor collections
        actor_collections = assemble_id_url(type=UrlType.Collections, base_url=actor_id, primary=None)
        await self.collections.remove_from_collection(collection_id=actor_collections, items=collection_id, access=AccessType.PRIVATE)

        # re-add the items to the new collection
        await self.collections.add_to_collection(collection_id=collection_id_new, items=collection_items_public, access=AccessType.PUBLIC)
        await self.collections.add_to_collection(collection_id=collection_id_new, items=collection_items_follower, access=AccessType.FOLLOWER)
        await self.collections.add_to_collection(collection_id=collection_id_new, items=collection_items_private, access=AccessType.PRIVATE)

        # add the new collection to the actor collections
        await self.collections.add_to_collection(collection_id=actor_collections, items=collection_id_new, add_only_once=True)

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """Process Update activity."""
        # Update object
        obj = activity.object
        if not isinstance(obj, APObject):
            raise ValidationError("UpdateHandler: actor.object is not of type APObject")
        obj.updated = datetime.now(timezone.utc)
        object_id = dereference_or_raise(obj, "id")
        actor_id = dereference_or_raise(activity, "actor")
        object_type = dereference_or_raise(obj, "type")
        self.logger.info(f"Updating object with type {object_type}")

        if CollectionType.__contains__(object_type):
            object_id_parsed = urlparse(object_id)
            object_id = urlunparse([object_id_parsed.scheme, object_id_parsed.netloc, object_id_parsed.path, None, None, None])
            target_id = dereference_or_raise(activity, "target")
            await self.update_collection(
                actor_id=actor_id,
                collection_id=object_id,
                collection_id_new=target_id,
            )
        else:
            # update object
            # Preserve immutable fields
            url_type = UrlType.from_object_type(object_type)
            existing = await self.resolver.try_resolve_from_storage(url=object_id, url_type=url_type)
            if not existing:
                raise ValidationError(f"Object with id {obj.id} does not exist locally.")
            existing = existing.serialize()
            for field in ["id", "type", "attributedTo", "published"]:
                if field in existing:
                    setattr(obj, field, existing[field])

            # Store updated object
            await self._store_updated_object(obj)

        # Deliver update to recipients
        await self._deliver_update(activity)

        return activity

    async def _store_updated_object(self, obj: APObject):
        assert obj.id, "Object does not have a proper id"
        if isinstance(obj, APPrivateKey):
            _ = await self.storage.key.upsert(id=obj.id, data=obj)
        elif isinstance(obj, APActor):
            _ = await self.storage.actor.upsert(id=obj.id, data=obj)
        else:
            _ = await self.storage.object.upsert(id=obj.id, data=obj)

    async def _deliver_update(self, activity: APActivity) -> None:
        """Deliver update to recipients."""
        obj = await self.resolver.resolve_object(activity.object)

        # Collect recipients
        recipients: list[ObjectId] = []

        # Add mentioned users
        if obj.tag:
            for tag in obj.tag:
                if isinstance(tag, APLink):
                    tag_href = tag.href
                else:
                    tag = await self.resolver.resolve_object(tag)
                    tag_href = getattr(tag, "href", None)
                if not tag_href:
                    self.logger.warning(f"Tag href is not a proper URL for activity {activity.serialize()}")
                    return
                if tag.type == "Mention":
                    recipients.append(tag_href)

        # Add followers if public
        actor_id = dereference(activity, key="actor")
        if actor_id:
            actor = await self.storage.actor.read(id=actor_id)
            if actor and actor.followers:
                if obj.visibility == "public":
                    recipients.append(actor.followers)

        # Deliver activity
        if recipients:
            _ = await self.delivery.deliver_activity(activity=activity, recipients=recipients)

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        raise HandlerError("Update in inbox is not implemented.")



from datetime import datetime, timezone
from logging import Logger

from phederation.collections.collections import CollectionManager
from phederation.models import APAccount, APActor, APCreate, ActorType
from phederation.models.collections import ValidCollection
from phederation.storage import StorageBackend
from phederation.utils import (
    ObjectId,
    UrlType,
    actor_id_from_username,
    collection_id_from_name,
    configure_logger,
)
from phederation.utils.base import AccessType, ObjectType, urljoin
from phederation.utils.exceptions import HandlerError, UserError
from phederation.utils.settings import PhedSettings

from .key_management import KeyManager


class ActorManager:

    def __init__(self, settings: PhedSettings, storage: StorageBackend, key_manager: KeyManager, collections: CollectionManager) -> None:
        self.storage: StorageBackend = storage
        self.key_manager: KeyManager = key_manager
        self.collections: CollectionManager = collections
        self.settings: PhedSettings = settings
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

    async def actor_from_storage(self, id: ObjectId):
        """Attempt to read an APActor object with the given actor id from storage.

        Note that access levels are disregarded here, it is a direct read from storage.

        Args:
            id (ObjectId): Actor id from the actor to read.

        Returns:
            APActor: An object, or None if not found.
        """
        return await self.storage.actor.read(id)

    async def _create_actor_collections(
        self,
        username: str,
        actor_id: ObjectId,
        actor_type: ActorType | ObjectType = ActorType.PERSON,
        autoaccept: bool = True,
        discoverable: bool = True,
    ) -> APActor:
        actor: APActor | None = await self.storage.actor.read(id=actor_id)

        if actor is None:
            # create the keys for the actor
            active_key = await self.key_manager.generate_key_pair(actor_id=actor_id)

            # create all collections
            inbox = collection_id_from_name(actor_id, name=ValidCollection.Inbox.value)
            outbox = collection_id_from_name(actor_id, name=ValidCollection.Outbox.value)
            followers = collection_id_from_name(actor_id, name=ValidCollection.Followers.value)
            following = collection_id_from_name(actor_id, name=ValidCollection.Following.value)
            blocks = collection_id_from_name(actor_id, name=ValidCollection.Blocks.value)
            liked = collection_id_from_name(actor_id, name=ValidCollection.Liked.value)
            objects = collection_id_from_name(actor_id, name=ValidCollection.Objects.value)
            lists = collection_id_from_name(actor_id, name=ValidCollection.Collections.value)

            endpoints: dict[str, ObjectId] | None = {}
            endpoints["id"] = actor_id + "#endpoints"
            if "media" in self.settings.federation.endpoints:
                endpoints["uploadMedia"] = urljoin(actor_id, "endpoints", "uploadMedia")
            if ValidCollection.Objects.value in self.settings.federation.endpoints:
                endpoints[ValidCollection.Objects.value] = objects
            if ValidCollection.Collections.value in self.settings.federation.endpoints:
                endpoints[ValidCollection.Collections.value] = lists
            if len(endpoints.keys()) == 0:
                endpoints = None

            if "sharedInbox" in self.settings.federation.endpoints:
                shared_inbox_url = urljoin(self.settings.domain.base_url, "/inbox")
            else:
                shared_inbox_url = None

            # TODO: make collection visibility selectable by actor. Right now, all is PUBLIC

            # generate APActor
            actor_obj = APActor(
                type=str(actor_type),
                id=actor_id,
                autoaccept=autoaccept,
                preferred_username=username,
                inbox=await self.collections.create_collection(collection_id=inbox, attributed_to=actor_id),
                outbox=await self.collections.create_collection(collection_id=outbox, attributed_to=actor_id),
                followers=await self.collections.create_collection(collection_id=followers, attributed_to=actor_id),
                following=await self.collections.create_collection(collection_id=following, attributed_to=actor_id),
                blocks=await self.collections.create_collection(collection_id=blocks, attributed_to=actor_id),
                liked=await self.collections.create_collection(collection_id=liked, attributed_to=actor_id),
                objects=await self.collections.create_collection(collection_id=objects, attributed_to=actor_id),
                collections=await self.collections.create_collection(collection_id=lists, attributed_to=actor_id),
                shared_inbox=shared_inbox_url,
                endpoints=endpoints,
                public_key=active_key.key_id,
                visibility=AccessType.PUBLIC.value,  # Important! Actors are always public (for webfinger), but may not be "discoverable".
                discoverable=discoverable,
            )

            _ = await self.storage.actor.create(data=actor_obj, raise_if_exists=True)
            actor = await self.storage.actor.read(id=actor_id)

            # also store in users collection. Make them private if not discoverable.
            collection_users = collection_id_from_name(id=self.settings.domain.base_url, name=UrlType.Users.value)
            await self.collections.add_to_collection(
                collection_id=collection_users, items=actor_id, access=AccessType.PUBLIC if actor_obj.discoverable else AccessType.PRIVATE
            )
        if not actor:
            raise ValueError("Actor could not be created.")
        return actor

    async def create_actor(self, activity: APCreate) -> APActor | None:
        """Tries to create an actor on the local instance, including a separate "account" for them."""
        if not activity.object:
            raise HandlerError("Activity object is None")
        if not isinstance(activity.object, APActor):
            raise HandlerError(f"Activity object type is not APActor (is {type(activity.object)})")
        actor: APActor = activity.object
        username = actor.preferred_username if actor.preferred_username else "<username>"
        actor_type = actor.type
        autoaccept = actor.autoaccept

        actor_id = actor_id_from_username(base_url=self.settings.domain.base_url, username=username)

        actor_types_to_create = [ActorType.PERSON, ActorType.GROUP, ActorType.SERVICE, ActorType.APPLICATION, ActorType.ORGANIZATION]
        if ActorType(actor_type) not in actor_types_to_create:
            raise HandlerError(f"Actor type must be in {actor_types_to_create}.")

        actor_storage = await self.storage.actor.read(id=actor_id)
        if actor_storage:
            self.logger.info(f"Tried to create actor '{actor_id}' but they already exist.")
            return None
        actor_created = await self._create_actor_collections(
            username=username, actor_id=actor_id, actor_type=ActorType(actor_type), autoaccept=autoaccept, discoverable=actor.discoverable
        )

        # create the account for the new user if it is a Person.
        if ActorType(actor_type) == ActorType.PERSON:
            account = APAccount(
                id=actor_id,
                username=username,
                fullname=actor.name,
                email=None,  # TODO: deal with submitting email for new accounts
                created_at=datetime.now(timezone.utc),
                updated_at=None,
                visibility=actor.visibility,
            )
            _ = await self.storage.account.create(data=account, raise_if_exists=False)

        self.logger.info(f"Created new user on the instance, including account: {actor_id}.")
        return actor_created

    async def create_actor_from_username(self, username: str):
        base_url = self.settings.domain.base_url
        create_actor_activity = APCreate(
            object=APActor(
                preferred_username=username,
                autoaccept=self.settings.federation.auto_accept_follows_default,
                type=ActorType.PERSON.value,
                inbox=urljoin(base_url, UrlType.Users.value, username, ValidCollection.Inbox.value),
                outbox=urljoin(base_url, UrlType.Users.value, username, ValidCollection.Outbox.value),
                visibility=AccessType.PUBLIC.value,
                discoverable=True,
            ),
            actor=actor_id_from_username(base_url=self.settings.domain.base_url, username=username),
        )
        _ = await self.create_actor(create_actor_activity)

    async def update_actor(self, update: APActor, actor_id: ObjectId):
        """Update an actor on this instance.
        If the actor is already in storage, updates their mutable information.

        Mutable fields for actors
         - preferred_username
         - autoaccept
        
        Args:
            update (APActor): The information for the update.
            actor_id (ObjectId): The id of the actor to update.

        Returns:
            APActor: the information of the updated actor, with the original .name and .id fields.
        
        Raises:
            UserError:  If the actor does not already exist in storage, or if the actor_id is not local to the instance (i.e. the actor is a remote actor).
        """
        if not self.settings.domain.is_local(actor_id):
            raise UserError(f"Actor {actor_id} is not a local actor")
        actor_original = await self.actor_from_storage(id=actor_id)
        if not actor_original:
            raise UserError(f"Actor {actor_id} does not exist in storage")
        
        # update mutable fields
        actor_original.preferred_username = update.preferred_username
        actor_original.autoaccept = update.autoaccept
        _ = await self.storage.actor.upsert(id=actor_id, data=actor_original)
        return actor_original

    async def delete_actor(self, actor_id: ObjectId):
        """Removes the actor with the given id.

        Removed
         - actor in storage
         - account of the actor in storage, if it exists
         - actor id from users collection

        Args:
            actor_id (ObjectId): _description_
        """
        actor_deleted = await self.storage.actor.delete(id=actor_id)

        if actor_deleted:
            _ = await self.storage.account.delete(id=actor_id)
            users_collection_id = collection_id_from_name(id=self.settings.domain.base_url, name=UrlType.Users.value)
            await self.collections.remove_from_collection(users_collection_id, items=actor_id, access=AccessType.PRIVATE)

    async def deactivate_actor(self, actor_id: ObjectId):
        """Deactivates the given actor (if they are an actor in storage).

        Args:
            actor_id (ObjectId): Id of the actor in storage.

        Returns:
            bool: True if the actor could be found, otherwise False.
        """
        actor = await self.storage.actor.read(id=actor_id)
        if not actor:
            return False
        actor.memorial = True
        actor.discoverable = False
        _ = await self.storage.actor.upsert(id=actor_id, data=actor)
        return True



"""
Base storage interface.
"""

from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, AsyncIterator
from logging import Logger
import os
from typing import Any

from dotenv import dotenv_values
from fastapi.datastructures import UploadFile
from pydantic import SecretStr

from phederation.models import CRUD
from phederation.models.activities import APActivity, APFollow
from phederation.models.actors import APAccount, APActor
from phederation.models.collections import APOrderedCollectionPage
from phederation.models.keys import APPrivateKey
from phederation.models.media import MediaMetadata
from phederation.models.objects import APCollectionItem, APObject
from phederation.models.proofs import DataIntegrityProof
from phederation.utils import NodeInfo, ObjectId
from phederation.utils.base import AccessType, ActivityPubBaseWithId
from phederation.utils.exceptions import StorageError, catch_exceptions
from phederation.utils.settings import PhedSettings, StorageProviderType


def fill_database_url_from_env(url: SecretStr, dotenv_path: str | None = None) -> SecretStr:
    """Fill in .env variables into the url, if necessary.

    Args:
        url (str): The database string with potentially unfilled variables (POSTGRES_PASSWORD) etc.
        dotenv_path (str|None): the path to the .env file to load from. Default: None (current directory).
    """
    url_secret = url.get_secret_value()
    settings = dotenv_values(dotenv_path=dotenv_path)

    if "mongodb" in url_secret:
        keys = [
            "MONGO_PHED_USER",
            "MONGO_PHED_PASSWORD",
            "MONGO_PHED_DEV_USER",
            "MONGO_PHED_DEV_PASSWORD",
            "MONGO_PHED_DB",
            "MONGO_PHED_HOST",
            "MONGO_PHED_PORT",
        ]
        if keys[0] in os.environ:  # if the key is already in the environment, do not use dotenv
            settings = {key: os.environ[key] for key in keys if key in os.environ}
        for key in keys:
            if key not in settings:
                raise EnvironmentError(f"Key {key} not in .env, but is required")
            value = settings.get(key)
            if value:
                url_secret = url_secret.replace(f"${{{key}}}", value)
    if "temporalio" in url_secret:
        keys = [
            "TEMPORAL_HOST",
            "TEMPORAL_PORT",
        ]
        if keys[0] in os.environ:  # if the key is already in the environment, do not use dotenv
            settings = {key: os.environ[key] for key in keys if key in os.environ}
        for key in keys:
            if key not in settings:
                raise EnvironmentError(f"Key {key} not in .env, but is required")
            value = settings.get(key)
            if value:
                url_secret = url_secret.replace(f"${{{key}}}", value)
            url_secret = url_secret.replace("temporalio://", "")
    return SecretStr(url_secret)


class StorageBackend(ABC):
    """Abstract storage backend."""

    # Register storage backends
    _providers: dict[StorageProviderType, type["StorageBackend"]] = {}

    actor: CRUD[APActor]
    account: CRUD[APAccount]
    object: CRUD[APObject]
    collection: CRUD[APOrderedCollectionPage]
    collection_items: CRUD[APCollectionItem]
    activity: CRUD[APActivity]
    follow: CRUD[APFollow]
    key: CRUD[APPrivateKey]
    proof: CRUD[DataIntegrityProof]
    node_info: CRUD[NodeInfo]

    database_url: SecretStr
    settings: PhedSettings
    logger: Logger

    def __init__(self, database_url: SecretStr, settings: PhedSettings) -> None:
        super().__init__()
        self.database_url = database_url
        self.settings = settings

    @classmethod
    def register_provider(cls, provider_name: StorageProviderType, provider_class: type["StorageBackend"]):
        """Register a storage provider."""
        cls._providers[provider_name] = provider_class

    @classmethod
    async def create(cls, settings: PhedSettings) -> "StorageBackend":
        """Create storage backend instance."""
        provider = StorageProviderType(settings.storage.provider)
        database_url = SecretStr(settings.storage.database.url)

        if provider not in cls._providers:
            raise StorageError(f"Unsupported storage provider: {provider}")

        provider_class = cls._providers[provider]
        return provider_class(database_url=database_url, settings=settings)

    @catch_exceptions(StorageError, "Failed to initialize storage")
    @abstractmethod
    async def initialize(self) -> None:
        """Initialize storage."""
        pass

    @catch_exceptions(StorageError, "Failed to reconnect to database")
    @abstractmethod
    async def reconnect(self) -> None:
        """Reconnect to the database. Useful if the async event loop changed, e.g. in delivery workers in different processes."""
        pass

    @abstractmethod
    def table[T: ActivityPubBaseWithId](self, model_class: type[T], name: str) -> CRUD[T]:
        """Get the table associated with the ActivityPubBaseWithId type."""

    @abstractmethod
    async def save(self,  document_id: ObjectId, content: UploadFile) -> ObjectId:
        """Save byte array data. Useful for media storage."""

    @abstractmethod
    async def load(self, file_id: ObjectId) -> AsyncIterator[bytes]:
        """Load byte array data. Useful for media storage."""

    @abstractmethod
    async def metadata(self, file_id: ObjectId) -> MediaMetadata:
        """Load metadata of a file, e.g. the file hash or content type. Useful for media storage."""

    @abstractmethod
    async def flush(self) -> None:
        """Flush storage to file, if possible."""
        pass

    @abstractmethod
    async def clear(self):
        """Delete the database. Useful for testing."""

    @abstractmethod
    async def close(self) -> None:
        """Close storage connection."""
        pass

    @abstractmethod
    async def is_following(self, follower: ObjectId, following: ObjectId, check_accepted: bool = False) -> bool:
        pass

    @abstractmethod
    async def paginate(self, collection_id: ObjectId, access: AccessType = AccessType.PUBLIC) -> AsyncGenerator[list[APCollectionItem], Any]:
        yield []

    @abstractmethod
    async def get_collection_page_items(
        self, collection_id: ObjectId, page: int = 1, access: AccessType = AccessType.PUBLIC
    ) -> list[APCollectionItem] | None:
        pass



"""
FastAPI integration implementation.
"""

from http import HTTPStatus
from logging import Logger
from typing import override
import uuid

from fastapi import Request
from fastapi.applications import FastAPI
from fastapi.encoders import jsonable_encoder
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from temporalio.client import Client

from phederation.actions.workflows import (
    ActivityDeliveryWorkflow,
    DeliveryWorkerTask,
    KeyRotationTask,
    UpdateNodeInfoTask,
)
from phederation.api import API_LATEST, create_api
from phederation.api.base import BaseAPI
from phederation.federation import DeliveryResult
from phederation.federation.middleware import ActivityPubMiddleware
from phederation.federation.server import ActivityPubServer
from phederation.federation.tasks import BackgroundTask, create_lifespan
from phederation.integration.base import BaseIntegration
from phederation.statistics.statistics import Statistics
from phederation.storage.tasks import StorageBackgroundTask
from phederation.utils.exceptions import (
    APIError,
    ActivityPubException,
    DeliveryError,
    IntegrationError,
    catch_exceptions,
)
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings
from phederation.utils.version import PHEDERATION_HEADERS, PHEDERATION_VERSION


class FastAPIIntegration(BaseIntegration):
    """FastAPI integration."""

    def __init__(self, settings: PhedSettings):
        super().__init__(settings)
        self.logger: Logger = configure_logger(__name__, self.settings.federation.logging_prefix)

        # Create the server with all components from the settings
        self.server: ActivityPubServer = ActivityPubServer(
            settings=self.settings, delivery_workflow_from_integration=FastAPIIntegration.delivery_workflow
        )
        self.background_tasks: list[BackgroundTask] = self._create_background_tasks()
        self.app: FastAPI = FastAPI(
            title="Phederation API",
            summary="A FastAPI implementation of ActivityPub for server2server and client2server communication.",
            description="The API allows all common routes in the ActivityPub protocol (user inbox/outbox, node info, webfinger, ...), and offers additional possibilities like user deletion/creation, media endpoints, and data integrity proofs.",
            version=PHEDERATION_VERSION,
            lifespan=create_lifespan(logger_prefix=self.settings.federation.logging_prefix, background_tasks=self.background_tasks),
        )
        self.app.add_middleware(
            middleware_class=CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["GET", "POST", "PUT", "DELETE"],
            allow_headers=["*"],
        )

        self.apis: dict[str, BaseAPI] = {}

    def _create_background_tasks(self) -> list[BackgroundTask]:
        if self.settings.federation.admin_mode:
            return []  # no background tasks for admin mode

        upgrade_node_info_task = UpdateNodeInfoTask(update_period=self.settings.federation.update_nodeinfo_period)
        delivery_task = DeliveryWorkerTask(update_period=self.settings.federation.delivery_queue_polling_period)
        key_rotation_task = KeyRotationTask(
            enabled=self.settings.security.key_rotation_enabled, update_period=self.settings.security.key_rotation_period
        )
        # worker_task = WorkerTask(
        #     workflows=[UpdateNodeInfoWorkflow, ActivityDeliveryWorkflow, KeyRotationWorkflow],
        #     activities=[
        #         upgrade_node_info_task.update_node_info,
        #         delivery_task.prepare_inboxes_for_delivery,
        #         delivery_task.deliver_activity,
        #         key_rotation_task.rotate_keys,
        #         key_rotation_task.announce_key_rotation
        #     ],
        # )
        return [
            upgrade_node_info_task,
            delivery_task,
            key_rotation_task,
            StorageBackgroundTask(),
        ]

    @override
    async def api(self, api_version: str | None = None) -> BaseAPI:
        if not api_version:
            api_version = API_LATEST

        api_version_major = f"{api_version.split('.')[0]}.X.X"

        if not api_version_major in self.apis:
            api = await create_api(api_version)
            await api.initialize(app=self.app, settings=self.settings, middleware=self.middleware, server=self.server)
            self.logger.info(f"Initialized API version {api.version()}.")

            api_version_major = api.version(major=True, minor=False, patch=False)
            self.apis[api_version_major] = api
        result = self.apis.get(api_version_major)
        if not result:
            raise APIError(f"Could not resolve api with version {api_version}")
        return result

    async def _setup_exception_handlers(self) -> None:

        @self.app.exception_handler(Exception)
        def handle_exception_error(request: Request, error: Exception):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
            error_type = str(type(error).__name__)
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            message = f"Not activitypub related"
            log_message = message

            if isinstance(error, ActivityPubException):
                ap_error: ActivityPubException = error
                message = ap_error.user_message or ap_error.default_message
                log_message = ap_error.message
                status_code = ap_error.status_code
            self.logger.error(f"FastAPIIntegration Exception ({error_type}; {log_message}). User-facing error: {error}.")

            headers = PHEDERATION_HEADERS
            headers["Host"] = self.settings.domain.netloc
            return JSONResponse(
                status_code=status_code,
                headers=headers,
                content=jsonable_encoder({"status": "error", "status_code": status_code, "type": str(type(error).__name__), "message": message}),
            )

    @staticmethod
    async def delivery_workflow(
        worker_client: Client | None,
        settings: PhedSettings,
    ):
        # TODO: refactoring. This method is just a crux to avoid importing the server in a cycle in Tasks.
        if not worker_client:
            raise DeliveryError("Worker client is not set")
        _ = await worker_client.start_workflow(
            ActivityDeliveryWorkflow.run,
            settings.federation,
            id=f"deliver-activity-{uuid.uuid4()}",
            task_queue=settings.worker.tasks_queue_name,
        )

        return DeliveryResult(status_code=HTTPStatus.CONTINUE)

    @catch_exceptions(IntegrationError, "Failed to initialize integration")
    @override
    async def initialize(self) -> None:
        """Initialize integration."""
        # this initializes all parameters of the ActivityPubServer.
        # It also clears the database, if specified in settings (e.g. for testing)
        await self.server.initialize()

        # after the server is initialized, we can create the instance actor(s)
        if not self.settings.federation.admin_mode:
            await self.server.create_instance_actors()

        # setup general exception handlers
        await self._setup_exception_handlers()

        # Create and initialize the statistics module logging middleware information
        self.statistics: Statistics = Statistics(
            app=self.app,
            settings=self.settings,
            storage=self.server.storage
        )
        await self.statistics.initialize()

        # Add the activitypub middleware
        self.middleware: ActivityPubMiddleware = ActivityPubMiddleware(
            settings=self.settings,
            statistics=self.statistics,
            server=self.server,
            app=self.app,
            signature_verifier=self.server.delivery.signature_verifier,
            rate_limiter=self.server.rate_limiter,
        )
        await self.middleware.initialize()

        # setup latest version of api
        _ = await self.api(api_version=API_LATEST)

        for task in self.background_tasks:
            self.logger.debug(f"FastAPI: initializing {type(task).__name__}")
            await task.initialize(server=self.server)

        if self.settings.federation.admin_mode:
            self.logger.info(f"FastAPI integration initialized, running in admin mode at {self.settings.domain.admin_url}")
        else:
            self.logger.info(f"FastAPI integration initialized, running at {self.settings.domain.base_url}")



from temporalio import workflow

with workflow.unsafe.imports_passed_through():
    from .api import *
    from .cache import *
    from .collections import *
    from .federation import *
    from .handlers import *
    from .integration import *
    from .models import *
    from .security import *
    from .statistics import *
    from .storage import *
    from .utils import *


"""
Announce activity handler implementation.
"""

from typing import override

from phederation.models import APPublicKey, dereference, object_from_type
from phederation.models.activities import APActivity
from phederation.models.objects import APObject, dereference_or_raise
from phederation.utils import ObjectId
from phederation.utils.base import AccessType, collection_id_from_name
from phederation.utils.exceptions import HandlerError, ValidationError

from .base import ActivityHandler


class AnnounceHandler(ActivityHandler):
    """Handle Announce (Boost/Reblog) activities."""

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Announce activity."""
        # Validate basic structure
        if activity.type != "Announce":
            raise ValidationError("Invalid activity type")

        if not activity.object:
            raise ValidationError("Missing object")

        if not activity.actor:
            raise ValidationError("Missing actor")

        # Validate object exists
        object_id = activity.object.id if isinstance(activity.object, APObject) else activity.object
        obj = await self.resolver.resolve_object(object_id)
        if not obj:
            raise ValidationError(f"Object not found: {object_id}")

        # Validate actor can announce
        actor = await self.resolver.resolve_actor(activity.actor)
        if not actor:
            raise ValidationError(f"Actor not found: {activity.actor}")

        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """Process Announce activity."""

        # Update object/actor shares collection
        object_id = dereference(activity, key="object")
        actor_id = dereference(activity, key="actor")
        if not object_id or not actor_id:
            raise HandlerError(f"AnnounceHandler: object_id or actor_id is None; object_id={object_id}, actor_id={actor_id}")

        access = AccessType.from_visibility(activity.visibility)
        await self._update_shares_collection(object_id=object_id, actor_id=actor_id, access=access)

        # Send notifications
        await self._notify_object_owner(object_id, activity)

        # Deliver to followers
        await self._deliver_to_followers(activity)

        return activity

    async def _update_shares_collection(self, object_id: ObjectId, actor_id: ObjectId, access: AccessType) -> None:
        """Update object's shares collection."""
        collection_shares = collection_id_from_name(id=object_id, name="shares")
        self.logger.debug(f"Adding actor {actor_id} to shares collection: {collection_shares}")
        # TODO: allow multiple shares if set in settings
        # TODO: it is not possible to "add only if not exists" here because of the storage session issue.
        _ = await self.collections.add_to_collection(collection_id=collection_shares, items=actor_id, access=access)

    async def _notify_object_owner(self, object_id: ObjectId, activity: APActivity) -> None:
        """Notify object owner about announce."""
        obj = await self.storage.object.read(id=object_id)
        if obj and obj.attributed_to:
            target_id = obj.attributed_to
            if isinstance(target_id, APObject):
                target_id = target_id.id
            if not target_id:
                self.logger.error(
                    f"AnnounceHandler: failed (error: target_id is Nont when trying to deliver Announce to actor '{target_id}. Activity: {activity}"
                )
                return

            # Inform the actor owning the object. This should create an addition to the "shares" collection of their object.
            # Only do this if the actor is not local, otherwise we will update twice.
            if not await self.delivery.check_is_local(actor_id=target_id):
                result = await self.delivery.deliver_to_actor_inbox(activity=activity, actor_id=target_id)
                if result.failed:
                    self.logger.error(
                        f"AnnounceHandler: failed (error: {result.error_message}) to deliver Announce to actor '{target_id}. Activity: {activity}"
                    )

    async def _deliver_to_followers(self, activity: APActivity) -> None:
        """Deliver announce to followers."""
        actor_id = dereference(activity, "actor")
        if not actor_id:
            raise HandlerError("AnnounceHandler, _deliver_to_followers: activity.actor is None")
        actor = await self.storage.actor.read(id=actor_id)
        if actor and actor.followers:
            _ = await self.delivery.deliver_activity(activity=activity, recipients=[actor.followers])

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        self.logger.debug(f"Processing Announce activity in inbox")

        # Update object/actor shares collection, if required. The object will not be a local one, but we may want to update local copies of it
        object_id = dereference_or_raise(activity, key="object")
        actor_id = dereference_or_raise(activity, key="actor")
        access = AccessType.from_visibility(activity.visibility)
        await self._update_shares_collection(object_id=object_id, actor_id=actor_id, access=access)

        if not isinstance(activity.object, APObject) and not isinstance(activity.object, dict):
            return
        obj = object_from_type(activity.object)
        if isinstance(obj, APPublicKey):
            self.logger.debug(f"Received new key in inbox, treating as key rotation", extra={"activity_id": activity.id})
            # at least check if the target is local to the instance.
            # Note that the *target* is not the actor with the new key here!
            # Key rotation always targets the instance actor (with "activity.bto"), but the "actor" is the one
            # on the remote instance that is rotating their keys.
            activity_target = dereference(activity.target, "id")
            activity_actor = dereference(activity.actor, "id")
            new_public_key = obj

            if not activity_target:
                self.logger.warning(f"Received a Create activity with key rotation data, but the activity.target is not set")
                return None
            if self.resolver.is_local(activity_target):
                self.logger.warning(
                    f"Received an Announce activity with key rotation data, but the activity.target actor {activity_target} is local to the instance (rotation should be federated). Discarding the new key"
                )
                return None
            if not activity_actor:
                self.logger.warning(f"Received a Create activity with key rotation data, but the activity.actor is not set")
                return None

            # store the new key in the actor list on the local instance
            actor_to_rotate = await self.resolver.resolve_actor(actor_id=activity_actor)
            if not actor_to_rotate.id:
                self.logger.warning(f"Received a Create activity with key rotation data, but the actor.id for {activity_actor} is None")
                return None

            new_key_id = dereference_or_raise(new_public_key, "id")
            actor_key_ids = actor_to_rotate.public_key_as_ids()
            if (
                (isinstance(actor_key_ids, list) and not (new_key_id in actor_key_ids))
                or (isinstance(actor_key_ids, str) and actor_key_ids != new_key_id)
                or actor_key_ids is None
            ):
                actor_to_rotate.add_public_key(new_public_key)
            _ = await self.storage.actor.upsert(id=actor_to_rotate.id, data=actor_to_rotate)

            # also store the key locally
            self.logger.debug(f"Storing new key locally, key_id={new_public_key.id}")
            _ = await self.key_manager.store_key(new_public_key)



"""
Collection handling implementation.
"""

from datetime import datetime, timezone
import math
from typing import override

from phederation.collections.collections import CollectionManager
from phederation.models import APOrderedCollectionPage, dereference
from phederation.storage.base import StorageBackend
from phederation.utils import ObjectId
from phederation.utils.base import AccessType
from phederation.utils.exceptions import CollectionError, catch_exceptions


class CollectionManagerDB(CollectionManager):
    """Manages ActivityPub collections, by storing any collection in the database."""

    def __init__(self, storage: StorageBackend, page_size: int = 20):
        super(CollectionManagerDB, self).__init__(storage=storage, page_size=page_size)

    @override
    async def initialize(self):
        pass

    @catch_exceptions(CollectionError, "Failed in create_collection")
    @override
    async def create_collection(
        self,
        collection_id: ObjectId,
        items: list[ObjectId] | None = None,
        attributed_to: ObjectId | None = None,
        access: AccessType = AccessType.PUBLIC,
    ) -> ObjectId:
        """
        Create a new collection.

        Args:
            collection_id: Collection id to create
            items: Initial collection items
            attributed_to: the actor or object this collection is attributed to.

        Returns:
            Collection ID
        """
        collection_id_check = await self.storage.collection.read(id=collection_id)
        if collection_id_check:
            self.logger.warning(f"Collection {collection_id} already exists")
            # raise CollectionError("Collection already exists")

        items = items or []

        total_items = len(items)
        n_pages = max(1, math.ceil(total_items / float(self.page_size)))
        items_split = [items[(k * self.page_size) : ((k + 1) * self.page_size)] for k in range(0, n_pages)]
        collections = [
            await self.create_new_page(
                collection_id=collection_id,
                page=page_number,
                n_pages=n_pages,
                attributed_to=attributed_to,
                total_items=len(items),
                items=items_split[page_number - 1],
            )
            for page_number in range(1, n_pages + 1)
        ]

        # create the general collection without the "page" id, linking to the first and last pages.
        # this collection does not have any items directly stored, but links to the pages.
        _ = await self.storage.collection.create(
            data=APOrderedCollectionPage(
                id=CollectionManager.collection_page(id=collection_id, page=None),
                attributed_to=attributed_to,
                total_items=len(items) if items else 0,
                ordered_items=None,
                published=datetime.now(timezone.utc),
                start_index=1,
                current=CollectionManager.collection_page(id=collection_id, page=1),
                first=CollectionManager.collection_page(id=collection_id, page=1),
                prev=None,
                next=None,
                last=CollectionManager.collection_page(id=collection_id, page=n_pages),
            )
        )

        # create all pages
        for collection in collections:
            _ = await self.storage.collection.create(data=collection)
        return collection_id

    @override
    async def pages(self, collection_id: ObjectId, attributed_to: ObjectId | None = None, access: AccessType = AccessType.PUBLIC):
        """Iterate over all pages of the collection.

        Args:
            collection_id (ObjectId): id of the collection. Can include a page, but will always start with the first item.

        Raises:
            CollectionError: If a part of the collection cannot be found.

        Yields:
            APOrderedCollectionPage: pages of the collection, one by one.
        """
        collection = await self.storage.collection.read(id=collection_id)
        first = dereference(collection, "first")
        if not collection or not first:
            raise CollectionError(f"Collection not found: {collection_id}")
        collection_id = first

        current_page = collection_id
        for page_number in range(1, 100):  # TODO: do not search more than 100 pages for safety; make this configurable in settings.
            current_id = dereference(current_page, key="id")
            if not current_id:
                raise CollectionError(f"Page {page_number} in {collection_id} did not have an id")
            current_page = await self.storage.collection.read(id=current_id)
            if not current_page:
                raise CollectionError(f"Page {page_number} in {collection_id} not found")
            if not current_page.id:
                raise CollectionError(f"Page {page_number} in {collection_id} did not have an id")

            yield current_page

            if current_page.id == current_page.last:
                break
            current_page = getattr(current_page, "next", None)

    @override
    async def contains(self, collection_id: ObjectId, item: ObjectId | None, access: AccessType = AccessType.PUBLIC) -> bool:
        """Check if the collection contains the given item.

        Args:
            collection_id (ObjectId): id of the collection
            item (ObjectId | None): item to check. Always returns False if item is None.

        Returns:
            bool: True if contained, False otherwise.
        """
        if not item:
            return False
        items_in_collection = await self.get_collection_items(collection_id=collection_id)
        return item in items_in_collection

    @catch_exceptions(CollectionError, "Failed in add_to_collection")
    @override
    async def add_to_collection(
        self, collection_id: ObjectId, items: ObjectId | list[ObjectId], access: AccessType = AccessType.PUBLIC, add_only_once: bool = False
    ) -> None:
        """
        Add items to collection.

        Args:
            collection_id: Collection ID
            items: Item(s) to add
        """
        if isinstance(items, str):
            items = [items]
        if len(items) == 0:
            return

        # create the collection if it does not already exist
        page_number = 1
        collection_id_with_page = CollectionManager.collection_page(id=collection_id, page=page_number)
        collection_page_one = await self.storage.collection.read(id=collection_id_with_page)
        if collection_page_one is None:
            self.logger.info(f"Could not find collection {collection_id_with_page}, creating it.")
            _ = await self.create_collection(collection_id=collection_id, items=items)
            return

        # collect all current pages
        all_pages = [p async for p in self.pages(collection_id=collection_id)]

        # loop until we find a collection with fewer than the page_size number of items
        collection_to_increase = None
        for page in all_pages:
            if len(page.items or page.ordered_items or []) < self.page_size:
                collection_to_increase = page
                break

        # if no incomplete page exists, create a new one
        if not collection_to_increase:
            collection_to_increase = await self.create_new_page(
                collection_id=collection_id,
                page=len(all_pages) + 1,
                n_pages=len(all_pages) + 1,
                attributed_to=dereference(collection_page_one, "attributed_to"),
                total_items=collection_page_one.total_items or 0,
                items=None,
            )

            # fix last pointer
            previous_last = dereference(collection_page_one, "last")
            for page in all_pages:
                if page.id == previous_last:
                    page.next = collection_to_increase.id
                    collection_to_increase.prev = previous_last
                page.last = collection_to_increase.id

            all_pages.append(collection_to_increase)

        # fix total items
        previous_total_items = collection_page_one.total_items or 0
        for page in all_pages:
            page.total_items = previous_total_items + len(items)

        # now we can actually add the items and update all pages
        previous_items = [dereference(item, "id") for item in collection_to_increase.ordered_items or []]
        previous_items = [item for item in previous_items if item]
        collection_to_increase.ordered_items = previous_items + items

        for page in all_pages:
            if page.id:
                _ = await self.storage.collection.upsert(id=page.id, data=page)

    @override
    async def remove_from_collection(self, collection_id: ObjectId, items: ObjectId | list[ObjectId], access: AccessType = AccessType.PUBLIC) -> None:
        """
        Remove items from collection.
        This deletes the entire collection, including all pages, and reforms it - meaning it is not very efficient.

        Args:
            collection_id: Collection ID
            items: Item(s) to remove
        """
        # fail silently if the collection does not exist
        if not (await self.collection_exists(collection_id=collection_id)):
            return
        
        collection = await self.get_collection_page(collection_id=collection_id, page=1)

        # only use lists
        if isinstance(items, str):
            items = [items]
        if len(items) == 0:
            return

        # get all items of the collection, remove the list
        items_in_collection = await self.get_collection_items(collection_id=collection_id)
        items_after_removal = list(set(items_in_collection) - set(items))

        # delete collection from storage
        await self.delete_collection(collection_id=collection_id)

        # create collection from items after removal
        attributed_to = dereference(collection, "attributed_to")
        _ = await self.create_collection(collection_id=collection_id, items=items_after_removal, attributed_to=attributed_to)

    @override
    async def delete_collection(self, collection_id: ObjectId):
        """Deletes the entire collection, including all pages, from storage.
        Items within the collection are not affected.

        Args:
            collection_id (ObjectId): The collection to remove.
        """
        # fail silently if the collection does not exist
        collection = await self.collection_exists(collection_id=collection_id)
        if not collection:
            return
        async for page in self.pages(collection_id=collection_id):
            if page.id:
                _ = await self.storage.collection.delete(page.id)

        _ = await self.storage.collection.delete(collection_id)

    @override
    async def collection_exists(self, collection_id: ObjectId, page: int | None = None, access: AccessType = AccessType.PUBLIC) -> bool:
        """Check if the collection exists.

        Args:
            collection_id (ObjectId): id of the collection to check.
            page (int | None, optional): Optional, only check if the given page exists. Defaults to None.

        Returns:
            APOrderedCollectionPage | None: The collection if it exists, or None.
        """
        if page:
            collection_id = CollectionManager.collection_page(id=collection_id, page=page)
        return await self.storage.collection.count(collection_id=collection_id, access=access) > 0

    @catch_exceptions(CollectionError, "Failed in get_collection_page")
    @override
    async def get_collection_page(
        self, collection_id: ObjectId, attributed_to: ObjectId | None = None, page: int = 1, access: AccessType = AccessType.PUBLIC
    ) -> APOrderedCollectionPage:
        """
        Get collection page.

        Args:
            collection_id: Collection ID
            page: Page number

        Returns:
            Collection page
        """
        # iterate over pages until we reach the desired one
        current_page = collection_id
        for page_number in range(1, page + 1):
            current_id = dereference(current_page, key="id")
            if not current_id:
                raise CollectionError(f"Page {page_number} in {collection_id} did not have an id")
            current_page = await self.storage.collection.read(id=current_id)
            current_page = getattr(current_page, "next", None) or getattr(current_page, "first", None)

        if isinstance(current_page, str):
            result_page = await self.storage.collection.read(id=current_page)
        else:
            result_page = current_page

        if not result_page:
            self.logger.error(f"Page {page} of collection {collection_id} is None. Page not found")
            raise CollectionError("Page not found")

        if not isinstance(result_page, APOrderedCollectionPage):
            self.logger.error(f"Page {page} of collection {collection_id} is not of correct type (is type {type(result_page)}).")
            raise CollectionError("Page not found")

        return result_page

    @override
    async def get_collection_items(self, collection_id: ObjectId, access: AccessType = AccessType.PUBLIC) -> list[ObjectId]:
        """Returns all item ids of all pages in the entire collection, not paginated.

        Args:
            collection_id (ObjectId): Id of the collection

        Returns:
            list[ObjectId]: All item ids.
        """
        # iterate over all pages, collecting items
        items: list[ObjectId] = []
        current_page = collection_id
        for page_number in range(1, 100):  # TODO: do not search more than 100 pages; make this configurable in settings.
            current_id = dereference(current_page, key="id")
            if not current_id:
                raise CollectionError(f"Page {page_number} in {collection_id} did not have an id")
            current_page = await self.storage.collection.read(id=current_id)
            if not current_page or not current_page.id:
                raise CollectionError(f"Page {page_number} in {collection_id} did not have an id")

            items_to_extend = current_page.items or current_page.ordered_items or []
            items.extend(items_to_extend)
            if current_page.id == current_page.last:
                break
            current_page = getattr(current_page, "next", None) or getattr(current_page, "first", None)
        return items



from abc import ABC, abstractmethod
from collections.abc import Awaitable
from typing import Any, Callable, Concatenate, cast

from pydantic import SecretStr

from phederation.utils.base import AccessType


class BaseCache(ABC):
    @abstractmethod
    def get(self, key: str) -> None | Any:
        pass

    @abstractmethod
    def set(self, key: str, value: Any):  # pyright: ignore[reportAny]
        pass

    @abstractmethod
    def delete(self, key: str):
        pass

    @abstractmethod
    def clear(
        self, prefix: str | None = None
    ):
        pass

    @abstractmethod
    def close(self):
        pass


class WithCache(ABC):
    """
    Specifies that the class has a cache property that can be accessed by the decorator.
    """

    cache: BaseCache | None = None


def with_cache(prefix: str):
    def with_cache_internal[**Params, SELF: WithCache, ReturnType](
        function: Callable[Concatenate[SELF, Params], Awaitable[ReturnType]],
    ) -> Callable[Concatenate[SELF, Params], Awaitable[ReturnType]]:
        async def wrapper_method(self: SELF, *args_with_cache: Params.args, **kwargs_with_cache: Params.kwargs):
            cache_or_none = self.cache
            args_joined = "-".join([str(arg) for arg in args_with_cache if isinstance(arg, str) or isinstance(arg, int)])
            sorted_keys = sorted(list(kwargs_with_cache.keys()))
            kwargs_joined_list: list[str] = []
            for arg in sorted_keys:
                value = kwargs_with_cache[arg]
                if isinstance(value, str) or isinstance(value, int) or isinstance(value, AccessType):
                    kwargs_joined_list.append(str(value))

                # this is important for security.
                # If an argument list contains a token or password, the result should not be cached.
                if isinstance(value, SecretStr):
                    cache_or_none = None
                    break
            kwargs_joined = "-".join(kwargs_joined_list)
            cache_key = "-".join([prefix, str(hash(args_joined)), str(hash(kwargs_joined))])
            if cache_or_none:
                if cached := cache_or_none.get(cache_key):
                    return cast(ReturnType, cached)
            result = await function(self, *args_with_cache, **kwargs_with_cache)
            if cache_or_none:
                cache_or_none.set(key=cache_key, value=result)
            return result

        return wrapper_method

    return with_cache_internal
